from .entity import *
from .event import *
from .flags import *
from .place_holder import *
from .preinstanced import *

from .icon import *
from .utils import *


__all__ = (
    *entity.__all__,
    *event.__all__,
    *flags.__all__,
    *place_holder.__all__,
    *preinstanced.__all__,
    
    *icon.__all__,
    *utils.__all__,
)



__all__ = ()

from .preinstanced import ScheduledEventEntityType


def guess_scheduled_event_entity_type_from_keyword_parameters(keyword_parameters):
    """
    Guesses scheduled event entity type from the given keyword parameters.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        Scheduled event metadata data.
    
    Returns
    -------
    metadata_type : `type<ScheduledEventEntityMetadataBase>`
    """
    if 'location' in keyword_parameters:
        metadata_type = ScheduledEventEntityType.location
    elif 'speaker_ids' in keyword_parameters:
        metadata_type = ScheduledEventEntityType.stage
    else:
        metadata_type = ScheduledEventEntityType.none
    
    return metadata_type



__all__ = ('ComponentMetadataSeparator',)

from scarletio import copy_docs

from .constants import SEPARATOR_SPACING_SIZE_DEFAULT
from .base import ComponentMetadataBase
from .fields import (
    parse_divider, parse_spacing_size, put_divider, put_spacing_size, validate_divider, validate_spacing_size
)


class ComponentMetadataSeparator(ComponentMetadataBase):
    """
    Represents the metadata of a separator component used inside of a form.
    
    Attributes
    ----------
    divider : `bool`
        Whether the separator should contain a divider.
    spacing_size : ``SeparatorSpacingSize``
        The separator's spacing's size.
    """
    __slots__ = ('divider', 'spacing_size')
    
    def __new__(
        cls,
        *,
        divider = ...,
        spacing_size = ...,
    ):
        """
        Creates a separator component metadata.
        
        Parameters
        ----------
        divider : `bool`, Optional (Keyword only)
            Whether the separator should contain a divider.
        spacing_size : ``SeparatorSpacingSize``, `int`, Optional (Keyword only)
            The separator's spacing's size.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # divider
        if divider is ...:
            divider = True
        else:
            divider = validate_divider(divider)
        
        # spacing_size
        if spacing_size is ...:
            spacing_size = SEPARATOR_SPACING_SIZE_DEFAULT
        else:
            spacing_size = validate_spacing_size(spacing_size)
        
        # Construct
        self = object.__new__(cls)
        self.divider = divider
        self.spacing_size = spacing_size
        return self
    
    
    @classmethod
    @copy_docs(ComponentMetadataBase.from_keyword_parameters)
    def from_keyword_parameters(cls, keyword_parameters):
        return cls(
            divider = keyword_parameters.pop('divider', ...),
            spacing_size = keyword_parameters.pop('spacing_size', ...),
        )
    
    
    @copy_docs(ComponentMetadataBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', type(self).__name__]
        
        # divider
        divider = self.divider
        if divider:
            repr_parts.append(' divider = ')
            repr_parts.append(repr(divider))
            
            field_added = True
        else:
            field_added = False
        
        # spacing_size
        spacing_size = self.spacing_size
        if (spacing_size is not SEPARATOR_SPACING_SIZE_DEFAULT):
            if field_added:
                repr_parts.append(',')
            
            repr_parts.append(' spacing_size = ')
            repr_parts.append(spacing_size.name)
            repr_parts.append(' ~ ')
            repr_parts.append(repr(spacing_size.value))
            
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(ComponentMetadataBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # divider
        hash_value ^= self.divider
        
        # spacing_size
        hash_value ^= self.spacing_size.value << 1
        
        return hash_value
    
    
    @copy_docs(ComponentMetadataBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        # divider
        if self.divider != other.divider:
            return False
        
        # spacing_size
        if self.spacing_size is not other.spacing_size:
            return False
        
        return True
    
    
    @classmethod
    @copy_docs(ComponentMetadataBase.from_data)
    def from_data(cls, data):
        self = object.__new__(cls)
        self.divider = parse_divider(data)
        self.spacing_size = parse_spacing_size(data)
        return self
    
    
    @copy_docs(ComponentMetadataBase.to_data)
    def to_data(self, *, defaults = False):
        data = {}
        
        put_divider(self.divider, data, defaults)
        put_spacing_size(self.spacing_size, data, defaults)
        
        return data
    
    
    @copy_docs(ComponentMetadataBase.copy)
    def copy(self):
        new = object.__new__(type(self))
        
        new.divider = self.divider
        new.spacing_size = self.spacing_size
        
        return new
    
    
    def copy_with(
        self, 
        *,
        divider = ...,
        spacing_size = ...,
    ):
        """
        Copies the separator component metadata with the given fields.
        
        Parameters
        ----------
        divider : `bool`, Optional (Keyword only)
            Whether the separator should contain a divider.
        spacing_size : ``SeparatorSpacingSize``, `int`, Optional (Keyword only)
            The separator's spacing's size.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # divider
        if divider is ...:
            divider = self.divider
        else:
            divider = validate_divider(divider)
        
        # spacing_size
        if spacing_size is ...:
            spacing_size = self.spacing_size
        else:
            spacing_size = validate_spacing_size(spacing_size)
        
        # Construct
        new = object.__new__(type(self))
        new.divider = divider
        new.spacing_size = spacing_size
        return new
    
    
    @copy_docs(ComponentMetadataBase.copy_with_keyword_parameters)
    def copy_with_keyword_parameters(self, keyword_parameters):
        return self.copy_with(
            divider = keyword_parameters.pop('divider', ...),
            spacing_size = keyword_parameters.pop('spacing_size', ...),
        )



__all__ = ()

from scarletio import include

from ...field_validators import default_entity_validator_factory, entity_validator_factory
from ...user import ClientUserBase, User, ZEROUSER, create_partial_user_from_id

from ..reaction.fields import parse_emoji, parse_type, put_emoji, put_type, validate_emoji, validate_type


Message = include('Message')

# user

def parse_user(data, guild_id = 0):
    """
    Parses out a user from the given reaction create or delete event data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Reaction event data.
    
    guild_id : `int` = `0`, Optional
        The guild's identifier where the event is from.
    
    Returns
    -------
    user : ``ClientUserBase``
    """
    if guild_id:
        try:
            guild_profile_data = data['member']
        except KeyError:
            pass
        else:
            user_data = guild_profile_data['user']
            return User.from_data(user_data, guild_profile_data, int(guild_id))
    
    return create_partial_user_from_id(int(data['user_id']))


def put_user(user, data, defaults, *, guild_id = 0):
    """
    Puts the given user's representation into the given reaction event data.
    
    Parameters
    ----------
    user : ``ClientUserBase``
        The user to serialize.
    data : `dict<str, object>`
        Reaction event data.
    defaults : `bool`
        Whether fields with their default values should be included as well.
    guild_id : `int` = `0`, Optional (Keyword only)
        The guild's identifier where the event is from.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    data['user_id'] = str(user.id)
    
    if guild_id:
        try:
            guild_profile = user.guild_profiles[guild_id]
        except KeyError:
            pass
        else:
            guild_profile_data = guild_profile.to_data(defaults = defaults, include_internals = True)
            guild_profile_data['user'] = user.to_data(defaults = defaults, include_internals = True)
            
            data['member'] = guild_profile_data
    
    return data


validate_user = default_entity_validator_factory('user', ClientUserBase, default = ZEROUSER)

# message

def parse_message(data):
    """
    Parses out a message from the given reaction create or delete event data.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        Reaction event data.
    
    Returns
    -------
    message : ``Message``
    """
    return Message._create_from_partial_data(data)


def put_message(message, data, defaults):
    """
    Puts the given message's representation into the given reaction event data.
    
    Parameters
    ----------
    message : ``Message``
        The message to serialize.
    data : `dict` of (`str`, `object`) items
        Reaction event data.
    defaults : `bool`
        Whether fields with their default values should be included as well.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    data['message_id'] = str(message.id)
    data['channel_id'] = str(message.channel_id)
    
    guild_id = message.guild_id
    if defaults or guild_id:
        if guild_id:
            guild_id = str(guild_id)
        else:
            guild_id = None
        
        data['guild_id'] = guild_id
    
    return data

validate_message = entity_validator_factory('message', NotImplemented, include = 'Message')



__all__ = ('require', )

from scarletio import get_last_module_frame

from ..exceptions import DoNotLoadPlugin


def require(*args, **kwargs):
    """
    Requires the given parameters.
    
    Parameters
    ----------
    *args : Parameters
        Required variable names.
    **kwargs : Keyword parameters
        Variables and their expected value / type.
    """
    module_frame = get_last_module_frame()
    if (module_frame is None):
        return
    
    module_globals = module_frame.f_globals
    
    for variable_name in args:
        if variable_name not in module_globals:
            raise DoNotLoadPlugin(variable_name)
    
    for variable_name, expected_value in kwargs.items():
        try:
            variable_value = module_globals[variable_name]
        except KeyError:
            raise DoNotLoadPlugin(variable_name) from None
        
        if variable_value is expected_value:
            continue
        
        if isinstance(expected_value, type) and isinstance(variable_value, expected_value):
            continue
        
        raise DoNotLoadPlugin(variable_name, variable_value, expected_value)



__all__ = ()

from ...field_validators import int_conditional_validator_factory

from .constants import CHANNELS_DEFAULT, FRAME_LENGTH_DEFAULT, SAMPLING_RATE_DEFAULT


validate_channels = int_conditional_validator_factory(
    'channels',
    CHANNELS_DEFAULT,
    lambda channels : channels >= 1,
    '>= 1',
)

validate_frame_length = int_conditional_validator_factory(
    'frame_length',
    FRAME_LENGTH_DEFAULT,
    lambda frame_length : frame_length >= 1,
    '>= 1',
)

validate_sampling_rate = int_conditional_validator_factory(
    'sampling_rate',
    SAMPLING_RATE_DEFAULT,
    lambda sampling_rate : sampling_rate >= 1,
    '>= 1',
)



__all__ = ()

from ...field_parsers import (
    bool_parser_factory, entity_id_parser_factory, force_string_parser_factory, nullable_object_array_parser_factory,
    preinstanced_parser_factory
)
from ...field_putters import (
    entity_id_putter_factory, force_bool_putter_factory, force_string_putter_factory,
    nullable_entity_array_putter_factory, preinstanced_putter_factory
)
from ...field_validators import (
    bool_validator_factory, entity_id_validator_factory, force_string_validator_factory,
    nullable_object_array_validator_factory, preinstanced_validator_factory
)

from ..onboarding_prompt_option import OnboardingPromptOption

from .preinstanced import OnboardingPromptType

# id

parse_id = entity_id_parser_factory('id')
put_id = entity_id_putter_factory('id')
validate_id = entity_id_validator_factory('option_id')

# in_onboarding

parse_in_onboarding = bool_parser_factory('in_onboarding', False)
put_in_onboarding = force_bool_putter_factory('in_onboarding')
validate_in_onboarding = bool_validator_factory('in_onboarding', False)

# name

parse_name = force_string_parser_factory('title')
put_name = force_string_putter_factory('title')
validate_name = force_string_validator_factory('name', 0, 1024)

# options

parse_options = nullable_object_array_parser_factory('options', OnboardingPromptOption)
put_options = nullable_entity_array_putter_factory('options', OnboardingPromptOption)
validate_options = nullable_object_array_validator_factory('options', OnboardingPromptOption)

# required

parse_required = bool_parser_factory('required', False)
put_required = force_bool_putter_factory('required')
validate_required = bool_validator_factory('required', False)

# single_select

parse_single_select = bool_parser_factory('single_select', False)
put_single_select = force_bool_putter_factory('single_select')
validate_single_select = bool_validator_factory('single_select', False)

# type

parse_type = preinstanced_parser_factory('type', OnboardingPromptType, OnboardingPromptType.multiple_choice)
put_type = preinstanced_putter_factory('type')
validate_type = preinstanced_validator_factory('prompt_type', OnboardingPromptType)



__all__ = ('ChannelMetadataPrivate',)

from scarletio import copy_docs

from ...permission.permission import PERMISSION_NONE, PERMISSION_PRIVATE, PERMISSION_PRIVATE_BOT

from .private_base import ChannelMetadataPrivateBase


class ChannelMetadataPrivate(ChannelMetadataPrivateBase):
    """
    Channel metadata for private channels.
    
    Attributes
    ----------
    users : `list` of ``ClientUserBase``
        The users in the channel.
    
    Class Attributes
    ----------------
    order_group: `int` = `0`
        The channel's order group used when sorting channels.
    """
    __slots__ = ()
    
    @copy_docs(ChannelMetadataPrivateBase._created)
    def _created(self, channel_entity, client, strong_cache):
        if (client is not None):
            users = self.users
            if client not in users:
                users.append(client)
                users.sort()
            
            if strong_cache and len(users) >= 2:
                user = users[0]
                if client is user:
                    user = users[1]
                
                client.private_channels[user.id] = channel_entity
    
    
    @copy_docs(ChannelMetadataPrivateBase._delete)
    def _delete(self, channel_entity, client):
        if (client is not None):
            users = self.users
            if len(users) >= 2:
                
                user = users[0]
                if client is user:
                    user = users[1]
                
                del client.private_channels[user.id]
    
    
    @property
    @copy_docs(ChannelMetadataPrivateBase.name)
    def name(self):
        users = self.users
        if len(users) >= 2:
            name = f'Direct Message {users[0].full_name} with {users[1].full_name}'
        else:
            name = f'Direct Message (partial)'
        return name
    
    
    @copy_docs(ChannelMetadataPrivateBase._get_permissions_for)
    def _get_permissions_for(self, channel_entity, user):
        if user in self.users:
            if user.bot:
                return PERMISSION_PRIVATE_BOT
            else:
                return PERMISSION_PRIVATE
            
        return PERMISSION_NONE



__all__ = ()

from ...bases import Icon
from ...scheduled_event import (
    PrivacyLevel, ScheduledEventEntityMetadataBase, ScheduledEventEntityType, ScheduledEventStatus
)
from ...scheduled_event.scheduled_event.fields import (
    validate_channel_id, validate_description, validate_end, validate_entity_id, validate_entity_type, validate_name,
    validate_privacy_level, validate_sku_ids, validate_start, validate_status
)
from ...scheduled_event.scheduled_event.scheduled_event import SCHEDULED_EVENT_IMAGE
from ...scheduled_event.scheduled_event_entity_metadata.fields import validate_location
from ...scheduled_event.scheduled_event_entity_metadata.utils import try_get_scheduled_event_metadata_type_from_data
from ...utils import datetime_to_timestamp, timestamp_to_datetime

from ..audit_log_entry_change_conversion import AuditLogEntryChangeConversion, AuditLogEntryChangeConversionGroup
from ..conversion_helpers.converters import (
    value_deserializer_description, value_deserializer_id, value_deserializer_ids, value_deserializer_name, value_serializer_description,
    value_serializer_id, value_serializer_ids, value_serializer_name
)


# ---- channel_id ----

CHANNEL_ID_CONVERSION = AuditLogEntryChangeConversion(
    ('channel_id',),
    'channel_id',
    value_deserializer = value_deserializer_id,
    value_serializer = value_serializer_id,
    value_validator = validate_channel_id,
)


# ---- description ----

DESCRIPTION_CONVERSION = AuditLogEntryChangeConversion(
    ('description',),
    'description',
    value_deserializer = value_deserializer_description,
    value_serializer = value_serializer_description,
    value_validator = validate_description,
)


# ---- entity_id ----

ENTITY_ID_CONVERSION = AuditLogEntryChangeConversion(
    ('entity_id',),
    'entity_id',
    value_deserializer = value_deserializer_id,
    value_serializer = value_serializer_id,
    value_validator = validate_entity_id,
)


# ---- entity_metadata ----

ENTITY_METADATA_CONVERSION = AuditLogEntryChangeConversion(
    ('entity_metadata',),
    'entity_metadata',
)

@ENTITY_METADATA_CONVERSION.set_value_deserializer
def entity_metadata_value_deserializer(value):
    if value is not None:
        metadata_type = try_get_scheduled_event_metadata_type_from_data(value)
        if metadata_type is None:
            value = None
        else:
            value = metadata_type.from_data(value)
    
    return value


@ENTITY_METADATA_CONVERSION.set_value_serializer
def entity_metadata_value_serializer(value):
    if value is not None:
        value = value.to_data(defaults = True)
    return value


@ENTITY_METADATA_CONVERSION.set_value_validator
def entity_metadata_value_validator(value):
    if value is None or isinstance(value, ScheduledEventEntityMetadataBase):
        return value
    
    raise TypeError(
        f'`entity_metadata` can be `None`, `{ScheduledEventEntityMetadataBase.__name__}`, '
        f'got {type(value).__name__}; {value!r}.'
    )


# ---- entity_type ----

ENTITY_TYPE_CONVERSION = AuditLogEntryChangeConversion(
    ('entity_type',),
    'entity_type',
    value_validator = validate_entity_type,
)


@ENTITY_TYPE_CONVERSION.set_value_deserializer
def entity_type_value_deserializer(value):
    return ScheduledEventEntityType(value)


@ENTITY_TYPE_CONVERSION.set_value_serializer
def entity_type_value_serializer(value):
    return value.value


# ---- image ----

IMAGE_CONVERSION = AuditLogEntryChangeConversion(
    ('image_hash',),
    'image',
    value_deserializer = Icon.from_base_16_hash,
    value_serializer = Icon.as_base_16_hash.fget,
    value_validator = SCHEDULED_EVENT_IMAGE.validate_icon,
)


# ---- location ----

LOCATION_CONVERSION = AuditLogEntryChangeConversion(
    ('location',),
    'location',
    value_deserializer = value_deserializer_description,
    value_serializer = value_serializer_description,
    value_validator = validate_location,
)

# ---- name ----

NAME_CONVERSION = AuditLogEntryChangeConversion(
    ('name',),
    'name',
    value_deserializer = value_deserializer_name,
    value_serializer = value_serializer_name,
    value_validator = validate_name,
)


# ---- privacy_level ----

PRIVACY_LEVEL_CONVERSION = AuditLogEntryChangeConversion(
    ('privacy_level',),
    'privacy_level',
    value_validator = validate_privacy_level,
)


@PRIVACY_LEVEL_CONVERSION.set_value_deserializer
def privacy_level_value_deserializer(value):
    return PrivacyLevel(value)


@PRIVACY_LEVEL_CONVERSION.set_value_serializer
def privacy_level_value_serializer(value):
    return value.value


# ---- end ----

END_CONVERSION = AuditLogEntryChangeConversion(
    ('scheduled_end_time',),
    'end',
    value_validator = validate_end,
)


# ---- start ----

START_CONVERSION = AuditLogEntryChangeConversion(
    ('scheduled_start_time',),
    'start',
    value_validator = validate_start,
)

# ----

@END_CONVERSION.set_value_deserializer
@START_CONVERSION.set_value_deserializer
def date_time_converter_get(value):
    if (value is not None):
        value = timestamp_to_datetime(value)
    return value

@END_CONVERSION.set_value_serializer
@START_CONVERSION.set_value_serializer
def date_time_converter_put(value):
    if (value is not None):
        value = datetime_to_timestamp(value)
    return value


# ---- sku_ids ----

SKU_IDS_CONVERSION = AuditLogEntryChangeConversion(
    ('sku_ids',),
    'sku_ids',
    value_deserializer = value_deserializer_ids,
    value_serializer = value_serializer_ids,
    value_validator = validate_sku_ids,
)


# ---- status ----

STATUS_CONVERSION = AuditLogEntryChangeConversion(
    ('status',),
    'status',
    value_validator = validate_status,
)


@STATUS_CONVERSION.set_value_deserializer
def status_value_deserializer(value):
    return ScheduledEventStatus(value)


@STATUS_CONVERSION.set_value_serializer
def status_value_serializer(value):
    return value.value


# ---- Construct ----

SCHEDULED_EVENT_CONVERSIONS = AuditLogEntryChangeConversionGroup(
    CHANNEL_ID_CONVERSION,
    DESCRIPTION_CONVERSION,
    ENTITY_ID_CONVERSION,
    ENTITY_METADATA_CONVERSION,
    ENTITY_TYPE_CONVERSION,
    IMAGE_CONVERSION,
    LOCATION_CONVERSION,
    NAME_CONVERSION,
    PRIVACY_LEVEL_CONVERSION,
    END_CONVERSION,
    START_CONVERSION,
    SKU_IDS_CONVERSION,
    STATUS_CONVERSION,
)



__all__ = ()

from ...field_parsers import (
    bool_parser_factory, entity_id_parser_factory, flag_parser_factory, force_string_parser_factory,
    nullable_date_time_parser_factory, nullable_string_parser_factory, preinstanced_array_parser_factory,
    preinstanced_parser_factory
)
from ...field_putters import (
    bool_optional_putter_factory, entity_id_putter_factory, flag_optional_putter_factory, force_string_putter_factory,
    nullable_date_time_optional_putter_factory, preinstanced_array_putter_factory, preinstanced_putter_factory,
    url_optional_putter_factory
)
from ...field_validators import (
    bool_validator_factory, entity_id_validator_factory, flag_validator_factory, force_string_validator_factory,
    nullable_date_time_validator_factory, preinstanced_array_validator_factory, preinstanced_validator_factory,
    url_optional_validator_factory
)

from ..application import Application

from .constants import NAME_LENGTH_MAX, NAME_LENGTH_MIN
from .flags import SKUFlag
from .preinstanced import SKUAccessType, SKUFeature, SKUType

# access_type

parse_access_type = preinstanced_parser_factory('access_type', SKUAccessType, SKUAccessType.none)
put_access_type = preinstanced_putter_factory('access_type')
validate_access_type = preinstanced_validator_factory('access_type', SKUAccessType)

# application_id

parse_application_id = entity_id_parser_factory('application_id')
put_application_id = entity_id_putter_factory('application_id')
validate_application_id = entity_id_validator_factory('application_id', Application)

# features

parse_features = preinstanced_array_parser_factory('features', SKUFeature)
put_features = preinstanced_array_putter_factory('features')
validate_features = preinstanced_array_validator_factory('features', SKUFeature)

# flags

parse_flags = flag_parser_factory('flags', SKUFlag)
put_flags = flag_optional_putter_factory('flags', SKUFlag())
validate_flags = flag_validator_factory('flags', SKUFlag)

# id

parse_id = entity_id_parser_factory('id')
put_id = entity_id_putter_factory('id')
validate_id = entity_id_validator_factory('sku_id')

# name

parse_name = force_string_parser_factory('name')
put_name = force_string_putter_factory('name')
validate_name = force_string_validator_factory('name', NAME_LENGTH_MIN, NAME_LENGTH_MAX)

# premium

parse_premium = bool_parser_factory('premium', False)
put_premium = bool_optional_putter_factory('premium', False)
validate_premium = bool_validator_factory('premium', False)

# release_at

parse_release_at = nullable_date_time_parser_factory('release_date')
put_release_at = nullable_date_time_optional_putter_factory('release_date')
validate_release_at = nullable_date_time_validator_factory('release_at')

# slug

parse_slug = nullable_string_parser_factory('slug')
put_slug = url_optional_putter_factory('slug')
validate_slug = url_optional_validator_factory('slug')

# type

parse_type = preinstanced_parser_factory('type', SKUType, SKUType.none)
put_type = preinstanced_putter_factory('type')
validate_type = preinstanced_validator_factory('sku_type', SKUType)



__all__ = ()

from .preinstanced import Locale


LOCALE_DEFAULT = Locale.english_us
Locale.INSTANCES[''] = LOCALE_DEFAULT


def build_locale_dictionary(dictionary):
    """
    Builds a locale dictionary where they keys are all ``Locale``-s.
    
    Parameters
    ----------
    dictionary : `None | dict<str, object>`
        The dictionary to process.
    
    Returns
    -------
    transformed : `None | dict<Locale, object>`
    """
    if (dictionary is not None) and dictionary:
        return {Locale(key): value for key, value in dictionary.items()}


def destroy_locale_dictionary(dictionary):
    """
    Builds a json serializable dictionary where they keys are all `str`-s.
    
    Parameters
    ----------
    dictionary : `None w dict<Locale | object>`
        The dictionary to process.
    
    Returns
    -------
    transformed : `None | dict<str, object>`
    """
    if dictionary is not None:
        return {key.value: value for key, value in dictionary.items()}


def hash_locale_dictionary(dictionary):
    """
    Hashes a locale dictionary where they keys are all ``Locale``-s.
    
    Parameters
    ----------
    dictionary : `dict<Locale, object>`
        The dictionary to process.
    
    Returns
    -------
    hash_value : `int`
    """
    hash_value = 0
    
    for key, value in dictionary.items():
        hash_value ^= hash(key.value) & hash(value)
    
    return hash_value



__all__ = ('OnboardingMode',)

from ...bases import Preinstance as P, PreinstancedBase


class OnboardingMode(PreinstancedBase, value_type = int):
    """
    Onboarding mode.
    
    Attributes
    ----------
    name : `str`
        The default name of the onboarding mode.
    
    value : `int`
        The discord side identifier value of the onboarding mode.
    
    Type Attributes
    ---------------
    Every predefined onboarding mode can be accessed as type attribute as well:
    
    +-------------------------------+---------------------------+-------+
    | Type attribute name           | name                      | value |
    +===============================+===========================+=======+
    | default                       | default                   | 0     |
    +-------------------------------+---------------------------+-------+
    | advanced                      | advanced                  | 1     |
    +-------------------------------+---------------------------+-------+
    """
    __slots__ = ()
    
    default = P(0, 'default')
    advanced = P(1, 'advanced')



__all__ = ('IntentFlag',)

from warnings import warn

from ...env import CACHE_PRESENCE

from ..bases import FlagBase, FlagDescriptor as F


INTENT_SHIFT_GUILDS = 0
INTENT_SHIFT_GUILD_USERS = 1
INTENT_SHIFT_GUILD_MODERATION = 2
INTENT_SHIFT_GUILD_EXPRESSIONS = 3
INTENT_SHIFT_GUILD_INTEGRATIONS = 4
INTENT_SHIFT_GUILD_WEBHOOKS = 5
INTENT_SHIFT_GUILD_INVITES = 6
INTENT_SHIFT_GUILD_VOICE_STATES = 7
INTENT_SHIFT_GUILD_PRESENCES = 8
INTENT_SHIFT_GUILD_MESSAGES = 9
INTENT_SHIFT_GUILD_REACTIONS = 10
INTENT_SHIFT_GUILD_TYPINGS = 11
INTENT_SHIFT_DIRECT_MESSAGES = 12
INTENT_SHIFT_DIRECT_REACTIONS = 13
INTENT_SHIFT_DIRECT_TYPINGS = 14
INTENT_SHIFT_MESSAGE_CONTENT = 15
INTENT_SHIFT_GUILD_SCHEDULED_EVENTS = 16
INTENT_SHIFT_AUTO_MODERATION_CONFIGURATION = 20
INTENT_SHIFT_AUTO_MODERATION_EXECUTION = 21
INTENT_SHIFT_GUILD_POLLS = 24
INTENT_SHIFT_DIRECT_POLLS = 25


INTENT_MASK_GUILDS = 1 << INTENT_SHIFT_GUILDS
INTENT_MASK_GUILD_USERS = 1 << INTENT_SHIFT_GUILD_USERS
INTENT_MASK_GUILD_MODERATION = 1 << INTENT_SHIFT_GUILD_MODERATION
INTENT_MASK_GUILD_EXPRESSIONS = 1 << INTENT_SHIFT_GUILD_EXPRESSIONS
INTENT_MASK_GUILD_INTEGRATIONS = 1 << INTENT_SHIFT_GUILD_INTEGRATIONS
INTENT_MASK_GUILD_WEBHOOKS = 1 << INTENT_SHIFT_GUILD_WEBHOOKS
INTENT_MASK_GUILD_INVITES = 1 << INTENT_SHIFT_GUILD_INVITES
INTENT_MASK_GUILD_VOICE_STATES = 1 << INTENT_SHIFT_GUILD_VOICE_STATES
INTENT_MASK_GUILD_PRESENCES = 1 << INTENT_SHIFT_GUILD_PRESENCES
INTENT_MASK_GUILD_MESSAGES = 1 << INTENT_SHIFT_GUILD_MESSAGES
INTENT_MASK_GUILD_REACTIONS = 1 << INTENT_SHIFT_GUILD_REACTIONS
INTENT_MASK_GUILD_TYPINGS = 1 << INTENT_SHIFT_GUILD_TYPINGS
INTENT_MASK_DIRECT_MESSAGES = 1 << INTENT_SHIFT_DIRECT_MESSAGES
INTENT_MASK_DIRECT_REACTIONS = 1 << INTENT_SHIFT_DIRECT_REACTIONS
INTENT_MASK_DIRECT_TYPINGS = 1 << INTENT_SHIFT_DIRECT_TYPINGS
INTENT_MASK_MESSAGE_CONTENT = 1 << INTENT_SHIFT_MESSAGE_CONTENT
INTENT_MASK_GUILD_SCHEDULED_EVENTS = 1 << INTENT_SHIFT_GUILD_SCHEDULED_EVENTS
INTENT_MASK_AUTO_MODERATION_CONFIGURATION = 1 << INTENT_SHIFT_AUTO_MODERATION_CONFIGURATION
INTENT_MASK_AUTO_MODERATION_EXECUTION = 1 << INTENT_SHIFT_AUTO_MODERATION_EXECUTION
INTENT_MASK_GUILD_POLLS = 1 << INTENT_SHIFT_GUILD_POLLS
INTENT_MASK_DIRECT_POLLS = 1 << INTENT_SHIFT_DIRECT_POLLS


INTENT_SHIFT_EVENTS = {
    INTENT_SHIFT_GUILDS: (
        'GUILD_CREATE',
        'GUILD_DELETE',
        'GUILD_UPDATE',
        'GUILD_ROLE_CREATE',
        'GUILD_ROLE_UPDATE',
        'GUILD_ROLE_DELETE',
        'CHANNEL_CREATE',
        'CHANNEL_UPDATE',
        'CHANNEL_DELETE',
        'CHANNEL_PINS_UPDATE',
        'STAGE_INSTANCE_CREATE',
        'STAGE_INSTANCE_UPDATE',
        'STAGE_INSTANCE_DELETE',
        'THREAD_CREATE',
        'THREAD_UPDATE',
        'THREAD_DELETE',
        'THREAD_LIST_SYNC',
        'THREAD_MEMBER_UPDATE',
        'THREAD_MEMBERS_UPDATE',
        'GUILD_APPLICATION_COMMAND_COUNTS_UPDATE',
        'GUILD_JOIN_REQUEST_CREATE',
        'GUILD_JOIN_REQUEST_DELETE',
        'GUILD_JOIN_REQUEST_UPDATE',
        'EMBEDDED_ACTIVITY_UPDATE',
        'EMBEDDED_ACTIVITY_UPDATE_V2',
        'CHANNEL_TOPIC_UPDATE',
        'VOICE_CHANNEL_STATUS_UPDATE',
    ),
    INTENT_SHIFT_GUILD_USERS: (
        'GUILD_MEMBER_ADD',
        'GUILD_MEMBER_UPDATE',
        'GUILD_MEMBER_REMOVE',
        'THREAD_MEMBERS_UPDATE',
    ),
    INTENT_SHIFT_GUILD_MODERATION: (
        'GUILD_AUDIT_LOG_ENTRY_CREATE',
        'GUILD_BAN_ADD',
        'GUILD_BAN_REMOVE',
    ),
    INTENT_SHIFT_GUILD_EXPRESSIONS: (
        'GUILD_EMOJIS_UPDATE',
        'GUILD_SOUNDBOARD_SOUND_CREATE',
        'GUILD_SOUNDBOARD_SOUND_DELETE',
        'GUILD_SOUNDBOARD_SOUND_UPDATE',
        'GUILD_SOUNDBOARD_SOUNDS_UPDATE',
        'GUILD_STICKERS_UPDATE',
        'SOUNDBOARD_SOUNDS',
    ),
    INTENT_SHIFT_GUILD_INTEGRATIONS: (
        'INTEGRATION_CREATE',
        'INTEGRATION_DELETE',
        'INTEGRATION_UPDATE',
        'GUILD_INTEGRATIONS_UPDATE',
    ),
    INTENT_SHIFT_GUILD_WEBHOOKS: (
        'WEBHOOKS_UPDATE',
    ),
    INTENT_SHIFT_GUILD_INVITES: (
        'INVITE_CREATE',
        'INVITE_DELETE',
    ),
    INTENT_SHIFT_GUILD_VOICE_STATES: (
        'VOICE_CHANNEL_EFFECT_SEND',
        'VOICE_STATE_UPDATE',
    ),
    INTENT_SHIFT_GUILD_PRESENCES: (
        'PRESENCE_UPDATE',
    ),
    INTENT_SHIFT_GUILD_MESSAGES: (
        'CHANNEL_PINS_UPDATE',
        'MESSAGE_CREATE',
        'MESSAGE_UPDATE',
        'MESSAGE_DELETE',
        'MESSAGE_DELETE_BULK', # Not listed by Discord, yayyyy
    ),
    INTENT_SHIFT_GUILD_REACTIONS: (
        'MESSAGE_REACTION_ADD',
        'MESSAGE_REACTION_REMOVE',
        'MESSAGE_REACTION_REMOVE_ALL',
        'MESSAGE_REACTION_REMOVE_EMOJI',
    ),
    INTENT_SHIFT_GUILD_TYPINGS: (
        'TYPING_START',
    ),
    INTENT_SHIFT_DIRECT_MESSAGES: (
        'CHANNEL_CREATE',
        'CHANNEL_PINS_UPDATE',
        'MESSAGE_CREATE',
        'MESSAGE_UPDATE',
        'MESSAGE_DELETE',
    ),
    INTENT_SHIFT_DIRECT_REACTIONS: (
        'MESSAGE_REACTION_ADD',
        'MESSAGE_REACTION_REMOVE',
        'MESSAGE_REACTION_REMOVE_ALL',
        'MESSAGE_REACTION_REMOVE_EMOJI',
    ),
    INTENT_SHIFT_DIRECT_TYPINGS: (
        'TYPING_START',
    ),
    INTENT_SHIFT_MESSAGE_CONTENT: (
    ),
    INTENT_SHIFT_GUILD_SCHEDULED_EVENTS: (
        'GUILD_SCHEDULED_EVENT_CREATE',
        'GUILD_SCHEDULED_EVENT_UPDATE',
        'GUILD_SCHEDULED_EVENT_DELETE',
        'GUILD_SCHEDULED_EVENT_USER_ADD',
        'GUILD_SCHEDULED_EVENT_USER_REMOVE',
    ),
    INTENT_SHIFT_AUTO_MODERATION_CONFIGURATION: (
        'AUTO_MODERATION_RULE_CREATE',
        'AUTO_MODERATION_RULE_DELETE',
        'AUTO_MODERATION_RULE_UPDATE',
    ),
    INTENT_SHIFT_AUTO_MODERATION_EXECUTION: (
        'AUTO_MODERATION_ACTION_EXECUTION',
    ),
    INTENT_SHIFT_GUILD_POLLS: (
        'MESSAGE_POLL_VOTE_ADD',
        'MESSAGE_POLL_VOTE_REMOVE',
    ),
    INTENT_SHIFT_DIRECT_POLLS: (
        'MESSAGE_POLL_VOTE_ADD',
        'MESSAGE_POLL_VOTE_REMOVE',
    ),
}

GLOBAL_INTENT_SHIFT_EVENTS = (
    'READY',
    'RESUMED',
    'USER_UPDATE',
    'CHANNEL_RECIPIENT_ADD', # User account only
    'CHANNEL_RECIPIENT_REMOVE', # User only
    'GUILD_MEMBERS_CHUNK',
    'VOICE_SERVER_UPDATE',
    'RELATIONSHIP_ADD', # User account only
    'RELATIONSHIP_REMOVE', # User account only
    'PRESENCES_REPLACE', # Empty / User account
    'USER_SETTINGS_UPDATE', # User account only
    'GIFT_CODE_UPDATE',
    'USER_ACHIEVEMENT_UPDATE', # User account only
    'MESSAGE_ACK', # User account only
    'SESSIONS_REPLACE', # User account only
    'INTERACTION_CREATE',
    'APPLICATION_COMMAND_CREATE',
    'APPLICATION_COMMAND_UPDATE',
    'APPLICATION_COMMAND_DELETE',
    'APPLICATION_COMMAND_PERMISSIONS_UPDATE',
    'USER_GUILD_SETTINGS_UPDATE', # User account only
    'CHANNEL_UNREAD_UPDATE', # User account only
    'GUILD_APPLICATION_COMMAND_INDEX_UPDATE', # Not documented
    'ENTITLEMENT_CREATE',
    'ENTITLEMENT_DELETE',
    'ENTITLEMENT_UPDATE',
    'SUBSCRIPTION_CREATE',
    'SUBSCRIPTION_DELETE',
    'SUBSCRIPTION_UPDATE',
)

INTENT_SHIFT_DEFAULT_EVENT = 255
INTENT_SHIFT_MISSING_EVENT = 254


DISPATCH_EVENT_TO_INTENTS = {}

def populate_dispatch_event_intents():
    from itertools import chain
    
    for intent_shift, event_names in chain(
        INTENT_SHIFT_EVENTS.items(), ((INTENT_SHIFT_DEFAULT_EVENT, GLOBAL_INTENT_SHIFT_EVENTS),)
    ):
        
        for event_name in event_names:
            try:
                intent_shifts = DISPATCH_EVENT_TO_INTENTS[event_name]
            except KeyError:
                intent_shifts = (intent_shift, )
            else:
                intent_shifts = (*intent_shifts, intent_shift, )
            
            DISPATCH_EVENT_TO_INTENTS[event_name] = intent_shifts

populate_dispatch_event_intents()
del populate_dispatch_event_intents


class IntentFlag(FlagBase):
    """
    An `int` subclass representing the intents to receive specific events. The wrapper picks these up as well and
    optimizes the dispatch events' events.
    
    Each flag specifies which parser's dispatch event is received from Discord. Not mentioned events do not depend
    on intent flags and they are expected to be received independently.
    
    +-----------------------------------------------+-------+-------------------------------+---------------------------------------------+
    | Intent flag position's                        | Shift | Intent name                   | Corresponding parser                        |
    | respective name                               | value |                               |                                             |
    +===============================================+=======+===============================+=============================================+
    | INTENT_SHIFT_GUILDS                           | 0     | guilds                        | GUILD_CREATE,                               |
    |                                               |       |                               | GUILD_DELETE,                               |
    |                                               |       |                               | GUILD_UPDATE,                               |
    |                                               |       |                               | GUILD_ROLE_CREATE,                          |
    |                                               |       |                               | GUILD_ROLE_UPDATE,                          |
    |                                               |       |                               | GUILD_ROLE_DELETE,                          |
    |                                               |       |                               | CHANNEL_CREATE,                             |
    |                                               |       |                               | CHANNEL_UPDATE,                             |
    |                                               |       |                               | CHANNEL_DELETE,                             |
    |                                               |       |                               | CHANNEL_PINS_UPDATE,                        |
    |                                               |       |                               | STAGE_INSTANCE_CREATE,                      |
    |                                               |       |                               | STAGE_INSTANCE_UPDATE,                      |
    |                                               |       |                               | STAGE_INSTANCE_DELETE,                      |
    |                                               |       |                               | THREAD_CREATE,                              |
    |                                               |       |                               | THREAD_UPDATE,                              |
    |                                               |       |                               | THREAD_DELETE,                              |
    |                                               |       |                               | THREAD_LIST_SYNC,                           |
    |                                               |       |                               | THREAD_MEMBER_UPDATE,                       |
    |                                               |       |                               | THREAD_MEMBERS_UPDATE,                      |
    |                                               |       |                               | GUILD_JOIN_REQUEST_CREATE,                  |
    |                                               |       |                               | GUILD_JOIN_REQUEST_DELETE,                  |
    |                                               |       |                               | GUILD_JOIN_REQUEST_UPDATE,                  |
    |                                               |       |                               | EMBEDDED_ACTIVITY_UPDATE,                   |
    |                                               |       |                               | EMBEDDED_ACTIVITY_UPDATE_V2,                |
    |                                               |       |                               | CHANNEL_TOPIC_UPDATE,                       |
    |                                               |       |                               | VOICE_CHANNEL_STATUS_UPDATE                 |
    +-----------------------------------------------+-------+-------------------------------+---------------------------------------------+
    | INTENT_SHIFT_GUILD_USERS                      | 1     | guild_users                   | GUILD_MEMBER_ADD,                           |
    |                                               |       |                               | GUILD_MEMBER_UPDATE,                        |
    |                                               |       |                               | GUILD_MEMBER_REMOVE,                        |
    |                                               |       |                               | THREAD_MEMBERS_UPDATE                       |
    +-----------------------------------------------+-------+-------------------------------+---------------------------------------------+
    | INTENT_SHIFT_GUILD_MODERATION                 | 2     | guild_bans                    | GUILD_AUDIT_LOG_ENTRY_CREATE,               |
    |                                               |       |                               | GUILD_BAN_ADD,                              |
    |                                               |       |                               | GUILD_BAN_REMOVE                            |
    +-----------------------------------------------+-------+-------------------------------+---------------------------------------------+
    | INTENT_SHIFT_GUILD_EXPRESSIONS                | 3     | guild_expressions             | GUILD_EMOJIS_UPDATE,                        |
    |                                               |       |                               | GUILD_SOUNDBOARD_SOUND_CREATE,              |
    |                                               |       |                               | GUILD_SOUNDBOARD_SOUND_DELETE,              |
    |                                               |       |                               | GUILD_SOUNDBOARD_SOUND_UPDATE,              |
    |                                               |       |                               | GUILD_SOUNDBOARD_SOUNDS_UPDATE,             |
    |                                               |       |                               | GUILD_STICKERS_UPDATE,                      |
    |                                               |       |                               | SOUNDBOARD_SOUNDS                           |
    +-----------------------------------------------+-------+-------------------------------+---------------------------------------------+
    | INTENT_SHIFT_GUILD_INTEGRATIONS               | 4     | guild_integrations            | INTEGRATION_CREATE,                         |
    |                                               |       |                               | INTEGRATION_DELETE,                         |
    |                                               |       |                               | INTEGRATION_UPDATE,                         |
    |                                               |       |                               | GUILD_INTEGRATIONS_UPDATE                   |
    +-----------------------------------------------+-------+-------------------------------+---------------------------------------------+
    | INTENT_SHIFT_GUILD_WEBHOOKS                   | 5     | guild_webhooks                | WEBHOOKS_UPDATE                             |
    +-----------------------------------------------+-------+-------------------------------+---------------------------------------------+
    | INTENT_SHIFT_GUILD_INVITES                    | 6     | guild_invites                 | INVITE_CREATE,                              |
    |                                               |       |                               | INVITE_DELETE                               |
    +-----------------------------------------------+-------+-------------------------------+---------------------------------------------+
    | INTENT_SHIFT_GUILD_VOICE_STATES               | 7     | guild_voice_states            | VOICE_CHANNEL_EFFECT_SEND,                  |
    |                                               |       |                               | VOICE_STATE_UPDATE                          |
    +-----------------------------------------------+-------+-------------------------------+---------------------------------------------+
    | INTENT_SHIFT_GUILD_PRESENCES                  | 8     | guild_presences               | PRESENCE_UPDATE                             |
    +-----------------------------------------------+-------+-------------------------------+---------------------------------------------+
    | INTENT_SHIFT_GUILD_MESSAGES                   | 9     | guild_messages                | CHANNEL_PINS_UPDATE,                        |
    |                                               |       |                               | MESSAGE_CREATE,                             |
    |                                               |       |                               | MESSAGE_UPDATE,                             |
    |                                               |       |                               | MESSAGE_DELETE,                             |
    |                                               |       |                               | MESSAGE_DELETE_BULK                         |
    +-----------------------------------------------+-------+-------------------------------+---------------------------------------------+
    | INTENT_SHIFT_GUILD_REACTIONS                  | 10    | guild_reactions               | MESSAGE_REACTION_ADD,                       |
    |                                               |       |                               | MESSAGE_REACTION_REMOVE,                    |
    |                                               |       |                               | MESSAGE_REACTION_REMOVE_ALL,                |
    |                                               |       |                               | MESSAGE_REACTION_REMOVE_EMOJI               |
    +-----------------------------------------------+-------+-------------------------------+---------------------------------------------+
    | INTENT_SHIFT_GUILD_TYPINGS                    | 11    | guild_typings                 | TYPING_START                                |
    +-----------------------------------------------+-------+-------------------------------+---------------------------------------------+
    | INTENT_SHIFT_DIRECT_MESSAGES                  | 12    | direct_messages               | CHANNEL_CREATE,                             |
    |                                               |       |                               | CHANNEL_PINS_UPDATE,                        |
    |                                               |       |                               | MESSAGE_CREATE,                             |
    |                                               |       |                               | MESSAGE_UPDATE,                             |
    |                                               |       |                               | MESSAGE_DELETE                              |
    +-----------------------------------------------+-------+-------------------------------+---------------------------------------------+
    | INTENT_SHIFT_DIRECT_REACTIONS                 | 13    | direct_reactions              | MESSAGE_REACTION_ADD,                       |
    |                                               |       |                               | MESSAGE_REACTION_REMOVE,                    |
    |                                               |       |                               | MESSAGE_REACTION_REMOVE_ALL,                |
    |                                               |       |                               | MESSAGE_REACTION_REMOVE_EMOJI               |
    +-----------------------------------------------+-------+-------------------------------+---------------------------------------------+
    | INTENT_SHIFT_DIRECT_TYPINGS                   | 14    | direct_typings                | TYPING_START                                |
    +-----------------------------------------------+-------+-------------------------------+---------------------------------------------+
    | INTENT_SHIFT_MESSAGE_CONTENT                  | 15    | message_content               | N/A                                         |
    +-----------------------------------------------+-------+-------------------------------+---------------------------------------------+
    | INTENT_SHIFT_GUILD_SCHEDULED_EVENTS           | 16    | guild_scheduled_events        | GUILD_APPLICATION_COMMAND_COUNTS_UPDATE,    |
    |                                               |       |                               | GUILD_SCHEDULED_EVENT_CREATE,               |
    |                                               |       |                               | GUILD_SCHEDULED_EVENT_UPDATE,               |
    |                                               |       |                               | GUILD_SCHEDULED_EVENT_DELETE,               |
    |                                               |       |                               | GUILD_SCHEDULED_EVENT_USER_ADD,             |
    |                                               |       |                               | GUILD_SCHEDULED_EVENT_USER_REMOVE           |
    +-----------------------------------------------+-------+-------------------------------+---------------------------------------------+
    | INTENT_SHIFT_AUTO_MODERATION_CONFIGURATION    | 20    | auto_moderation_configuration | AUTO_MODERATION_RULE_CREATE                 |
    |                                               |       |                               | AUTO_MODERATION_RULE_DELETE                 |
    |                                               |       |                               | AUTO_MODERATION_RULE_UPDATE                 |
    +-----------------------------------------------+-------+-------------------------------+---------------------------------------------+
    | INTENT_SHIFT_AUTO_MODERATION_EXECUTION        | 21    | auto_moderation_execution     | AUTO_MODERATION_ACTION_EXECUTION            |
    +-----------------------------------------------+-------+-------------------------------+---------------------------------------------+
    | INTENT_SHIFT_GUILD_POLLS                      | 24    | guild_polls                   | MESSAGE_POLL_VOTE_ADD,                      |
    |                                               |       |                               | MESSAGE_POLL_VOTE_REMOVE                    |
    +-----------------------------------------------+-------+-------------------------------+---------------------------------------------+
    | INTENT_SHIFT_DIRECT_POLLS                     | 25    | direct_polls                  | MESSAGE_POLL_VOTE_ADD,                      |
    |                                               |       |                               | MESSAGE_POLL_VOTE_REMOVE                    |
    +-----------------------------------------------+-------+-------------------------------+---------------------------------------------+
    """
    guilds = F(INTENT_SHIFT_GUILDS)
    guild_users = F(INTENT_SHIFT_GUILD_USERS)
    guild_moderation = F(INTENT_SHIFT_GUILD_MODERATION)
    guild_expressions = F(INTENT_SHIFT_GUILD_EXPRESSIONS)
    guild_integrations = F(INTENT_SHIFT_GUILD_INTEGRATIONS)
    guild_webhooks = F(INTENT_SHIFT_GUILD_WEBHOOKS)
    guild_invites = F(INTENT_SHIFT_GUILD_INVITES)
    guild_voice_states = F(INTENT_SHIFT_GUILD_VOICE_STATES)
    guild_presences = F(INTENT_SHIFT_GUILD_PRESENCES)
    guild_messages = F(INTENT_SHIFT_GUILD_MESSAGES)
    guild_reactions = F(INTENT_SHIFT_GUILD_REACTIONS)
    guild_typings = F(INTENT_SHIFT_GUILD_TYPINGS)
    direct_messages = F(INTENT_SHIFT_DIRECT_MESSAGES)
    direct_reactions = F(INTENT_SHIFT_DIRECT_REACTIONS)
    direct_typings = F(INTENT_SHIFT_DIRECT_TYPINGS)
    message_content = F(INTENT_SHIFT_MESSAGE_CONTENT)
    guild_scheduled_events = F(INTENT_SHIFT_GUILD_SCHEDULED_EVENTS)
    auto_moderation_configuration = F(INTENT_SHIFT_AUTO_MODERATION_CONFIGURATION)
    auto_moderation_execution = F(INTENT_SHIFT_AUTO_MODERATION_EXECUTION)
    guild_polls = F(INTENT_SHIFT_GUILD_POLLS)
    direct_polls = F(INTENT_SHIFT_DIRECT_POLLS)
    
    
    def __new__(cls, integer = -1):
        """
        Creates a new ``IntentFlag`` from the passed `integer`. If any invalid intent flag is passed, those
        will be removed. If the wrapper is started up without presence caching, then `.guild_presences` will be
        set to `False` by default.
        
        Parameters
        ----------
        integer : `int` = `-1`, Optional
            The value what will be converted ``IntentFlag``. If not passed or passed as a negative value,
            then returns an ``IntentFlag`` what contains all the enabled flags.
        
        Raises
        ------
        TypeError
            If `integer` was not passed as `int`.
        
        Notes
        -----
        The default created intent flags contain the privileged gateway intents, so if you have those disabled, or
        if those are not allowed for you, then make sure, you specify them.
        """
        if not isinstance(integer, int):
            raise TypeError(
                f'{cls.__name__} expected `int`, got {type(integer).__name__}; {integer!r}.'
            )
        
        intent_flag = 0
        if integer < 0:
            for value in cls.__shifts__.values():
                intent_flag = intent_flag | (1 << value)
            
            # If presence cache is disabled, disable presence updates
            if not CACHE_PRESENCE:
                intent_flag = intent_flag ^ (1 << INTENT_SHIFT_GUILD_PRESENCES)
        else:
            for value in cls.__shifts__.values():
                if (integer >> value) & 1:
                    intent_flag = intent_flag | (1 << value)
            
            # If presence cache is disabled, disable presence updates
            if not CACHE_PRESENCE:
                if intent_flag & INTENT_MASK_GUILD_PRESENCES:
                    intent_flag = intent_flag^INTENT_MASK_GUILD_PRESENCES
        
        return int.__new__(cls, intent_flag)
    
    
    def iterate_parser_names(self):
        """
        Yields every parser's name, what the intent flag allows to be received.
        
        This method is a generator.
        
        Yields
        ------
        parser_name : `str`
        """
        for shift in self.__shifts__.values():
            if (self >> shift) & 1:
                yield from INTENT_SHIFT_EVENTS[shift]
        
        yield from GLOBAL_INTENT_SHIFT_EVENTS
    
    
    def allow_guilds(self):
        warn(
            f'`{type(self).__name__}.allow_guilds` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.guilds.mask)
    
    
    def allow_guild_users(self):
        warn(
            f'`{type(self).__name__}.allow_guild_users` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.guild_users.mask)
    
    
    def allow_guild_moderation(self):
        warn(
            f'`{type(self).__name__}.allow_guild_moderation` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.guild_moderation.mask)
    
    
    def allow_guild_expressions(self):
        warn(
            f'`{type(self).__name__}.allow_guild_expressions` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.guild_expressions.mask)
    
    
    def allow_guild_integrations(self):
        warn(
            f'`{type(self).__name__}.allow_guild_integrations` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.guild_integrations.mask)
    
    
    def allow_guild_webhooks(self):
        warn(
            f'`{type(self).__name__}.allow_guild_webhooks` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.guild_webhooks.mask)
    
    
    def allow_guild_invites(self):
        warn(
            f'`{type(self).__name__}.allow_guild_invites` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.guild_invites.mask)
    
    
    def allow_guild_voice_states(self):
        warn(
            f'`{type(self).__name__}.allow_guild_voice_states` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.guild_voice_states.mask)
    
    
    def allow_guild_presences(self):
        warn(
            f'`{type(self).__name__}.allow_guild_presences` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.guild_presences.mask)
    
    
    def allow_guild_messages(self):
        warn(
            f'`{type(self).__name__}.allow_guild_messages` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.guild_messages.mask)
    
    
    def allow_guild_reactions(self):
        warn(
            f'`{type(self).__name__}.allow_guild_reactions` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.guild_reactions.mask)
    
    
    def allow_guild_typings(self):
        warn(
            f'`{type(self).__name__}.allow_guild_typings` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.guild_typings.mask)
    
    
    def allow_direct_messages(self):
        warn(
            f'`{type(self).__name__}.allow_direct_messages` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.direct_messages.mask)
    
    
    def allow_direct_reactions(self):
        warn(
            f'`{type(self).__name__}.allow_direct_reactions` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.direct_reactions.mask)
    
    
    def allow_direct_typings(self):
        warn(
            f'`{type(self).__name__}.allow_direct_typings` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.direct_typings.mask)
    
    
    def allow_message_content(self):
        warn(
            f'`{type(self).__name__}.allow_message_content` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.message_content.mask)
    
    
    def allow_guild_scheduled_events(self):
        warn(
            f'`{type(self).__name__}.allow_guild_scheduled_events` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.guild_scheduled_events.mask)
    
    
    def allow_auto_moderation_configuration(self):
        warn(
            f'`{type(self).__name__}.allow_auto_moderation_configuration` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.auto_moderation_configuration.mask)
    
    
    def allow_auto_moderation_execution(self):
        warn(
            f'`{type(self).__name__}.allow_auto_moderation_execution` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.auto_moderation_execution.mask)
    
    
    def allow_guild_polls(self):
        warn(
            f'`{type(self).__name__}.allow_guild_polls` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.guild_polls.mask)
    
    
    def allow_direct_polls(self):
        warn(
            f'`{type(self).__name__}.allow_direct_polls` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.direct_polls.mask)
    
    
    def deny_guilds(self):
        warn(
            f'`{type(self).__name__}.deny_guilds` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.guilds.mask)
    
    
    def deny_guild_users(self):
        warn(
            f'`{type(self).__name__}.deny_guild_users` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.guild_users.mask)
    
    
    def deny_guild_moderation(self):
        warn(
            f'`{type(self).__name__}.deny_guild_moderation` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.guild_moderation.mask)
    
    
    def deny_guild_expressions(self):
        warn(
            f'`{type(self).__name__}.deny_guild_expressions` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.guild_expressions.mask)
    
    
    def deny_guild_integrations(self):
        warn(
            f'`{type(self).__name__}.deny_guild_integrations` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.guild_integrations.mask)
    
    
    def deny_guild_webhooks(self):
        warn(
            f'`{type(self).__name__}.deny_guild_webhooks` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.guild_webhooks.mask)
    
    
    def deny_guild_invites(self):
        warn(
            f'`{type(self).__name__}.deny_guild_invites` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.guild_invites.mask)
    
    
    def deny_guild_voice_states(self):
        warn(
            f'`{type(self).__name__}.deny_guild_voice_states` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.guild_voice_states.mask)
    
    
    def deny_guild_presences(self):
        warn(
            f'`{type(self).__name__}.deny_guild_presences` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.guild_presences.mask)
    
    
    def deny_guild_messages(self):
        warn(
            f'`{type(self).__name__}.deny_guild_messages` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.guild_messages.mask)
    
    
    def deny_guild_reactions(self):
        warn(
            f'`{type(self).__name__}.deny_guild_reactions` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.guild_reactions.mask)
    
    
    def deny_guild_typings(self):
        warn(
            f'`{type(self).__name__}.deny_guild_typings` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.guild_typings.mask)
    
    
    def deny_direct_messages(self):
        warn(
            f'`{type(self).__name__}.deny_direct_messages` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.direct_messages.mask)
    
    
    def deny_direct_reactions(self):
        warn(
            f'`{type(self).__name__}.deny_direct_reactions` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.direct_reactions.mask)
    
    
    def deny_direct_typings(self):
        warn(
            f'`{type(self).__name__}.deny_direct_typings` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.direct_typings.mask)
    
    
    def deny_message_content(self):
        warn(
            f'`{type(self).__name__}.deny_message_content` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.message_content.mask)
    
    
    def deny_guild_scheduled_events(self):
        warn(
            f'`{type(self).__name__}.deny_guild_scheduled_events` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.guild_scheduled_events.mask)
    
    
    def deny_auto_moderation_configuration(self):
        warn(
            f'`{type(self).__name__}.deny_auto_moderation_configuration` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.auto_moderation_configuration.mask)
    
    
    def deny_auto_moderation_execution(self):
        warn(
            f'`{type(self).__name__}.deny_auto_moderation_execution` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.auto_moderation_execution.mask)
    
    
    def deny_guild_polls(self):
        warn(
            f'`{type(self).__name__}.deny_guild_polls` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.guild_polls.mask)
    
    
    def deny_direct_polls(self):
        warn(
            f'`{type(self).__name__}.deny_direct_polls` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.direct_polls.mask)



"""
Hata is an asynchronous Discord API wrapper built on top of scarletio. It is designed to be easy to use, with also
providing rich API offering everything what an advanced developer might need.

Named after Hata no Kokoro from Touhou Project.

Why hata
--------

- Multiple simultaneous clients

    Hata can run multiple clients from the same instance without sacrificing performance.

- Performant
    
    Fast concurrent code using async/await syntax, cache control, PyPy support and more!

- Newest API features
    
    Whatever Discord decides to release/update/break Hata will support it natively in no time!

- 100% Python

    Completely relies on Python! Easy to read, easy to understand, easy to code.

Usage
-----

```py
from hata import Client, Guild, wait_for_interruption

GUILD = Guild.precreate(guild_id)

Seija = Client('TOKEN', extensions = ['slash'])

@Seija.events
async def ready(client):
    print(f'{client:f} logged in.')

@Seija.interactions(guild = GUILD)
async def ping():
    \"\"\"ping-pong\"\"\"
    return 'pong'

Seija.start()

wait_for_interruption()
```

Hata leaves the main thread free, `client.start()` blocks it only till the client logs in (or fails it), although you
can still use the `start_clients()` function, what as it says, starts up all the non-running clients parallelly.

Sometimes leaving the main thread might cause problems when trying to shut down the bot(s). At this case, you might
want to use `wait_for_interruption()`, which disconnects the clients gracefully and closes the event loop on keyboard
interrupt.

We got some `tutorials:https://www.astil.dev/project/hata/guides/` as well, please check them too!
"""
__version__ = '1.3.76'

# First import env, so if exception occurs we do not load the whole library.
from .env import *

from .discord import *
from .ext import *
from .utils import *


__all__ = (
    *discord.__all__,
    *env.__all__,
    *ext.__all__,
    *utils.__all__,
)

# Additional imports

import sys

from scarletio import check_satisfaction

from .utils.module_deprecation import get_deprecation_function

# Check whether every export is satisfied

check_satisfaction()

# Setup deprecations

__getattr__ = get_deprecation_function()

# Setup tests

if 'vampytest' in sys.modules:
    from . import vampytest_config



__all__ = ('is_allowed_mentions_valid', 'parse_allowed_mentions')

from scarletio import include

from ..role import Role
from ..user import UserBase

from .constants import STATE_ALLOW_REPLIED_USER_FALSE, STATE_ALLOW_REPLIED_USER_NONE, STATE_ALLOW_REPLIED_USER_TRUE


AllowedMentionProxy = include('AllowedMentionProxy')


def is_allowed_mentions_element_valid(element):
    """
    Returns whether the given allowed mentions element is valid.
    
    Parameters
    ----------
    element : `object`
        The element to check.
    
    Returns
    -------
    is_valid : `bool`
    """
    if isinstance(element, str):
        return element in {'everyone', 'users', 'roles', 'replied_user', '!replied_user'}
    
    if isinstance(element, UserBase):
        return True
    
    if isinstance(element, Role):
        return True
    
    return False


def is_allowed_mentions_valid(allowed_mentions):
    """
    Returns whether the given value is valid as an `allowed_mentions` parameter.
    
    Parameters
    ----------
    allowed_mentions : `None`,  `str`, ``UserBase``, ``Role``, ``AllowedMentionProxy``, (`list`, `tuple`, `set`) of \
            (`str`, ``UserBase``, ``Role`` )
        Which user or role can the message ping (or everyone).
    
    Returns
    -------
    is_valid : `bool`
    """
    if allowed_mentions is None:
        return True
    
    if isinstance(allowed_mentions, AllowedMentionProxy):
        return True
    
    if isinstance(allowed_mentions, (list, set, tuple)):
        return all(is_allowed_mentions_element_valid(element) for element in allowed_mentions)
    
    return False


def parse_allowed_mentions(allowed_mentions):
    """
    If `allowed_mentions` is passed as `None`, then returns a `dict`, what will cause all mentions to be disabled.
    
    If passed as an `iterable`, then it's elements will be checked. They can be either type `str`
    (any value from `('everyone', 'users', 'roles')`), ``UserBase``, ``Role``-s.
    
    Passing `everyone` will allow the message to mention `@everyone` (permissions can overwrite this behaviour).
    
    Passing `'users'` will allow the message to mention all the users, meanwhile passing ``UserBase``-s.
    allow to mentioned the respective users. Using `users` and ``UserBase``-s. is mutually exclusive,
    and the wrapper will register only `users` to avoid getting ``DiscordException``.
    
    `'roles'` and ``Role``-s. follow the same rules as `'users'` and the ``UserBase``-s.
    
    By passing `'!replied_user'` you can disable mentioning the replied user, or by passing`'replied_user'` you can
    re-enable mentioning the replied user.
    
    Parameters
    ----------
    allowed_mentions : `None`,  `str`, ``UserBase``, ``Role``, ``AllowedMentionProxy``, (`list`, `tuple`, `set`) of \
            (`str`, ``UserBase``, ``Role`` )
        Which user or role can the message ping (or everyone).
    
    Returns
    -------
    allowed_mentions : `dict` of (`str`, `object`) items
    
    Raises
    ------
    TypeError
        If `allowed_mentions` contains an element of invalid type.
    ValueError
        If `allowed_mentions` contains en element of correct type, but an invalid value.
    """
    if (allowed_mentions is None):
        return {'parse': []}
    
    if isinstance(allowed_mentions, AllowedMentionProxy):
        return allowed_mentions.to_data()
    
    if isinstance(allowed_mentions, list):
        if (not allowed_mentions):
            return {'parse': []}
    
    elif isinstance(allowed_mentions, (set, tuple)):
        if (not allowed_mentions):
            return {'parse': []}
        
        allowed_mentions = [*allowed_mentions]
    
    else:
        allowed_mentions = [allowed_mentions]
    
    allow_replied_user = STATE_ALLOW_REPLIED_USER_NONE
    allow_everyone = 0
    allow_users = 0
    allow_roles = 0
    
    allowed_users = None
    allowed_roles = None
    
    for element in allowed_mentions:
        if isinstance(element, str):
            if element == '!replied_user':
                allow_replied_user = STATE_ALLOW_REPLIED_USER_FALSE
                continue
            
            if element == 'replied_user':
                allow_replied_user = STATE_ALLOW_REPLIED_USER_TRUE
                continue
            
            if element == 'everyone':
                allow_everyone = 1
                continue
            
            if element == 'users':
                allow_users = 1
                continue
            
            if element == 'roles':
                allow_roles = 1
                continue
            
            raise ValueError(
                f'`allowed_mentions` contains a not valid `str` element: `{element!r}`. `str` '
                f'elements can be any of: (\'everyone\', \'users\', \'roles\', \'replied_user\', '
                f'\'!replied_user\').'
            )
        
        if isinstance(element, UserBase):
            if allowed_users is None:
                allowed_users = []
            
            allowed_users.append(str(element.id))
            continue
        
        if isinstance(element, Role):
            if allowed_roles is None:
                allowed_roles = []
            
            allowed_roles.append(str(element.id))
            continue
        
        raise TypeError(
            f'`allowed_mentions` can contain `str`, `{Role.__name__}`, `{UserBase.__name__}` elements, got '
            f' {type(element).__name__}; {element!r}; allowed_mentions = {allowed_mentions!r}.'
        )
    
    
    result = {}
    parse_all_of = None
    
    if allow_replied_user != STATE_ALLOW_REPLIED_USER_NONE:
        result['replied_user'] = (allow_replied_user == STATE_ALLOW_REPLIED_USER_TRUE)
    
    if allow_everyone:
        if parse_all_of is None:
            parse_all_of = []
            result['parse'] = parse_all_of
        
        parse_all_of.append('everyone')
    
    if allow_users:
        if parse_all_of is None:
            parse_all_of = []
            result['parse'] = parse_all_of
        
        parse_all_of.append('users')
    else:
        if (allowed_users is not None):
            result['users'] = allowed_users
    
    if allow_roles:
        if parse_all_of is None:
            parse_all_of = []
            result['parse'] = parse_all_of
        
        parse_all_of.append('roles')
    else:
        if (allowed_roles is not None):
            result['roles'] = allowed_roles
    
    return result



__all__ = ('OnboardingPromptOption',)

from ...bases import DiscordEntity
from ...channel import ChannelType, create_partial_channel_from_id
from ...precreate_helpers import process_precreate_parameters_and_raise_extra
from ...role import create_partial_role_from_id

from .fields import (
    parse_channel_ids, parse_description, parse_emoji, parse_id, parse_name, parse_role_ids, put_channel_ids,
    put_description, put_emoji, put_id, put_name, put_role_ids, validate_channel_ids,
    validate_description, validate_emoji, validate_id, validate_name, validate_role_ids
)


PRECREATE_FIELDS = {
    'channel_ids': ('channel_ids', validate_channel_ids),
    'channels': ('channel_ids', validate_channel_ids),
    'description': ('description', validate_description),
    'emoji': ('emoji', validate_emoji),
    'name': ('name', validate_name),
    'role_ids': ('role_ids', validate_role_ids),
    'roles': ('role_ids', validate_role_ids),
}


class OnboardingPromptOption(DiscordEntity):
    """
    Option of an onboarding prompt.
    
    Attributes
    ----------
    channel_ids : `None`, `tuple` of `int`
        The channels' identifier opted into when this option is selected.
    description : `None`, `str`
        The option's description
    emoji : `None`, ``Emoji``
        The emoji of the option.
    id : `int`
        the option's identifier.
    name : `str`
        The option's name.
    role_ids : `None`, `tuple` of `int`
        The roles' identifiers assigned to the user when this option is selected.
    """
    __slots__ = ('channel_ids', 'description', 'emoji', 'name', 'role_ids')
    
    def __new__(cls, *, channel_ids = ..., description = ..., emoji = ..., name = ..., role_ids = ...):
        """
        Creates an onboarding prompt option instance from the given parameters.
        
        Parameters
        ----------
        channel_ids : `None`, `iterable` of (`int` ``Channel``), Optional (Keyword only)
            The channels' identifier opted into when this option is selected.
        description : `None`, `str`, Optional (Keyword only)
            The option's description
        emoji : `None`, `str`, Optional (Keyword only)
            The emoji of the option.
        name : `str`, Optional (Keyword only)
            The option's name.
        role_ids : `None`, `iterable` of (`int` ``Role``), Optional (Keyword only)
            The roles' identifiers assigned to the user when this option is selected.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # channel_ids
        if channel_ids is ...:
            channel_ids = None
        else:
            channel_ids = validate_channel_ids(channel_ids)
        
        # description
        if description is ...:
            description = None
        else:
            description = validate_description(description)
        
        # emoji
        if emoji is ...:
            emoji = None
        else:
            emoji = validate_emoji(emoji)
        
        # name
        if name is ...:
            name = ''
        else:
            name = validate_name(name)
        
        # role_ids
        if role_ids is ...:
            role_ids = None
        else:
            role_ids = validate_role_ids(role_ids)
        
        self = object.__new__(cls)
        self.channel_ids = channel_ids
        self.description = description
        self.emoji = emoji
        self.id = 0
        self.name = name
        self.role_ids = role_ids
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new onboarding prompt option from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Onboarding prompt option data.
        
        Returns
        -------
        new : `instance<cls>`
        """
        self = object.__new__(cls)
        self.channel_ids = parse_channel_ids(data)
        self.description = parse_description(data)
        self.emoji = parse_emoji(data)
        self.id = parse_id(data)
        self.name = parse_name(data)
        self.role_ids = parse_role_ids(data)
        return self
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Converts the onboarding prompt option to json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields with their default value should be included as well.
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        data = {}
        put_channel_ids(self.channel_ids, data, defaults)
        put_description(self.description, data, defaults)
        put_emoji(self.emoji, data, defaults)
        put_name(self.name, data, defaults)
        put_role_ids(self.role_ids, data, defaults)
        
        if include_internals:
            put_id(self.id, data, defaults)
        
        return data
    
    
    def __repr__(self):
        """Returns the onboarding prompt option's representation."""
        repr_parts = [
            '<',
            self.__class__.__name__,
        ]
        
        option_id = self.id
        if option_id:
            repr_parts.append(' id = ')
            repr_parts.append(repr(option_id))
            repr_parts.append(',')
        
        repr_parts.append(' name = ')
        repr_parts.append(repr(self.name))
        
        description = self.description
        if (description is not None):
            repr_parts.append(', description = ')
            repr_parts.append(repr(description))
        
        emoji = self.emoji
        if (emoji is not None):
            repr_parts.append(', emoji = ')
            repr_parts.append(repr(emoji))
        
        channel_ids = self.channel_ids
        if (channel_ids is not None):
            repr_parts.append(', channel_ids = [')
            
            index = 0
            length = len(channel_ids)
            
            while True:
                channel_id = channel_ids[index]
                index += 1
                
                repr_parts.append(repr(channel_id))
                
                if index == length:
                    break
                
                repr_parts.append(', ')
            
            repr_parts.append(']')
        
        role_ids = self.role_ids
        if (role_ids is not None):
            repr_parts.append(', role_ids = [')
            
            index = 0
            length = len(role_ids)
            
            while True:
                role_id = role_ids[index]
                index += 1
                
                repr_parts.append(repr(role_id))
                
                if index == length:
                    break
                
                repr_parts.append(', ')
            
            repr_parts.append(']')
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two onboarding prompt options are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal_same_type(other)
    
    
    def __ne__(self, other):
        """Returns whether the two onboarding prompt options are not equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return not self._is_equal_same_type(other)
    
    
    def _is_equal_same_type(self, other):
        """
        Returns whether the two onboarding prompt options are equal.
        
        Parameters
        ----------
        other : `instance<type<self>>`
            The other instance.
        
        Returns
        -------
        are_equal : `bool`
        """
        if self.channel_ids != other.channel_ids:
            return False
        
        if self.description != other.description:
            return False
        
        if self.emoji != other.emoji:
            return False
        
        # id
        # Ignore it
        
        if self.name != other.name:
            return False
        
        if self.role_ids != other.role_ids:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the onboarding prompt option's hash value."""
        hash_value = 0
        
        channel_ids = self.channel_ids
        if (channel_ids is not None):
            hash_value ^= len(channel_ids)
            
            for channel_id in channel_ids:
                hash_value ^= channel_id
        
        description = self.description
        if (description is not None):
            hash_value ^= hash(description)
        
        emoji = self.emoji
        if (emoji is not None):
            hash_value ^= hash(emoji)
        
        # id
        # ignore it
        
        name = self.name
        if (name is not None):
            if name != description:
                hash_value ^= hash(name)
        
        role_ids = self.role_ids
        if (role_ids is not None):
            hash_value ^= len(role_ids) << 4
            
            for role_id in role_ids:
                hash_value ^= role_id
        
        return hash_value
    
    
    @classmethod
    def precreate(
        cls,
        option_id,
        **keyword_parameters,
    ):
        """
        Precreates an onboarding prompt option. Since they are not cached, this method just a ``.__new__`` alternative.
        
        Parameters
        ----------
        option_id : `int`
            The option's identifier.
        
        **keyword_parameters : Keyword parameters
            Additional parameters defining how the option's fields should be set.
        
        Other Parameters
        ----------------
        channel_ids : `None`, `iterable` of (`int` ``Channel``), Optional (Keyword only)
            The channels' identifier opted into when this option is selected.
        channel_ids : `None`, `iterable` of (`int` ``Channel``), Optional (Keyword only)
            Alternative of `channel_ids`.
        description : `None`, `str`, Optional (Keyword only)
            The option's description
        emoji : `None`, `str`, Optional (Keyword only)
            The emoji of the option.
        name : `str`, Optional (Keyword only)
            The option's name.
        role_ids : `None`, `iterable` of (`int` ``Role``), Optional (Keyword only)
            The roles' identifiers assigned to the user when this option is selected.
        roles : `None`, `iterable` of (`int` ``Role``), Optional (Keyword only)
            Alternative of `role_ids`.
        
        Returns
        -------
        self : `instance<type<self>>`
        """
        option_id = validate_id(option_id)
        
        if keyword_parameters:
            processed = process_precreate_parameters_and_raise_extra(keyword_parameters, PRECREATE_FIELDS)
        else:
            processed = None
        
        # Construct
        
        self = object.__new__(cls)
        self.channel_ids = None
        self.description = None
        self.emoji = None
        self.id = option_id
        self.name = ''
        self.role_ids = None
        
        if (processed is not None):
            for item in processed:
                setattr(self, *item)
        
        return self
    
    
    def copy(self):
        """
        Copies the onboarding prompt option.
        
        Returns
        -------
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        
        channel_ids = self.channel_ids
        if (channel_ids is not None):
            channel_ids = (*channel_ids,)
        new.channel_ids = channel_ids
        
        new.description = self.description
        new.emoji = self.emoji
        new.id = 0
        new.name = self.name
        
        role_ids = self.role_ids
        if (role_ids is not None):
            role_ids = (*role_ids,)
        new.role_ids = role_ids
        
        return new
    
    
    def copy_with(self, *, channel_ids = ..., description = ..., emoji = ..., name = ..., role_ids = ...):
        """
        Copies the onboarding prompt option with the given fields.
        
        Parameters
        ----------
        channel_ids : `None`, `iterable` of (`int` ``Channel``), Optional (Keyword only)
            The channels' identifier opted into when this option is selected.
        description : `None`, `str`, Optional (Keyword only)
            The option's description
        emoji : `None`, `str`, Optional (Keyword only)
            The emoji of the option.
        name : `str`, Optional (Keyword only)
            The option's name.
        role_ids : `None`, `iterable` of (`int` ``Role``), Optional (Keyword only)
            The roles' identifiers assigned to the user when this option is selected.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # channel_ids
        if channel_ids is ...:
            channel_ids = self.channel_ids
            if (channel_ids is not None):
                channel_ids = (*channel_ids,)
        else:
            channel_ids = validate_channel_ids(channel_ids)
        
        # description
        if description is ...:
            description = self.description
        else:
            description = validate_description(description)
        
        # emoji
        if emoji is ...:
            emoji = self.emoji
        else:
            emoji = validate_emoji(emoji)
        
        # name
        if name is ...:
            name = self.name
        else:
            name = validate_name(name)
        
        # role_ids
        if role_ids is ...:
            role_ids = self.role_ids
            if (role_ids is not None):
                role_ids = (*role_ids,)
        else:
            role_ids = validate_role_ids(role_ids)
        
        new = object.__new__(type(self))
        new.channel_ids = channel_ids
        new.description = description
        new.emoji = emoji
        new.id = 0
        new.name = name
        new.role_ids = role_ids
        return new
    
    
    @property
    def roles(self):
        """
        Returns the roles that are assigned to the user when this option is selected.
        
        Returns
        -------
        roles : `None`, `tuple` of ``Role``
        """
        role_ids = self.role_ids
        if role_ids is None:
            roles = None
        else:
            roles = (*sorted(
                (create_partial_role_from_id(role_id) for role_id in role_ids),
            ),)
        
        return roles
    
    
    def iter_role_ids(self):
        """
        Iterates over the roles' identifiers assigned to the user when this option is selected.
        
        This method is an iterable generator.
        
        Yields
        ------
        role_id : `int`
        """
        role_ids = self.role_ids
        if (role_ids is not None):
            yield from role_ids
    
    
    def iter_roles(self):
        """
        Iterates over the roles that are assigned to the user when this option is selected.
        Not like ``.roles``, this will not sort the roles of the emoji based on their position, instead uses the
        default ordering (id).
        
        This method is an iterable generator.
        
        Yields
        ------
        role : ``Role``
        """
        for role_id in self.iter_role_ids():
            yield create_partial_role_from_id(role_id)
    
    
    @property
    def channels(self):
        """
        Returns the channels that opted into when this option is selected.
        
        Returns
        -------
        channels : `None`, `tuple` of ``Channel``
        """
        channel_ids = self.channel_ids
        if channel_ids is None:
            channels = None
        else:
            channels = (*(
                create_partial_channel_from_id(channel_id, ChannelType.unknown, 0) for channel_id in channel_ids
            ),)
        
        return channels
    
    
    def iter_channel_ids(self):
        """
        Iterates over the channels' identifiers that opted into when this option is selected.
        
        This method is an iterable generator.
        
        Yields
        ------
        channel_id : `int`
        """
        channel_ids = self.channel_ids
        if (channel_ids is not None):
            yield from channel_ids
    
    
    def iter_channels(self):
        """
        Iterates over the channels that opted into when this option is selected.
        
        This method is an iterable generator.
        
        Yields
        ------
        channel : ``Channel``
        """
        for channel_id in self.iter_channel_ids():
            yield create_partial_channel_from_id(channel_id, ChannelType.unknown, 0)



__all__ = ()

from ...field_parsers import flag_parser_factory, preinstanced_array_parser_factory
from ...field_putters import preinstanced_array_putter_factory, string_flag_putter_factory
from ...field_validators import flag_validator_factory, preinstanced_array_validator_factory
from ...oauth2 import Oauth2Scope
from ...permission import Permission
from ...permission.constants import PERMISSION_KEY

# permissions

parse_permissions = flag_parser_factory(PERMISSION_KEY, Permission)
put_permissions = string_flag_putter_factory(PERMISSION_KEY)
validate_permissions = flag_validator_factory('permissions', Permission)

# scopes

parse_scopes = preinstanced_array_parser_factory('scopes', Oauth2Scope)
put_scopes = preinstanced_array_putter_factory('scopes')
validate_scopes = preinstanced_array_validator_factory('scopes', Oauth2Scope)



__all__ = ('ChannelMetadataGuildThreadAnnouncements',)

from .guild_thread_base import ChannelMetadataGuildThreadBase


class ChannelMetadataGuildThreadAnnouncements(ChannelMetadataGuildThreadBase):
    """
    Guild channel metadata of announcement threads.
    
    Attributes
    ----------
    _created_at : `None`, `datetime`
        When the channel was created.
    _cache_permission : `None`, `dict` of (`int`, ``Permission``) items
        A `user_id` to ``Permission`` relation mapping for caching permissions. Defaults to `None`.
    archived : `bool`
        Whether the thread s archived.
    archived_at : `None`, `datetime`
        When the thread's archive status was last changed.
    auto_archive_after : `int`
        Duration in seconds to automatically archive the thread after recent activity. Can be one of: `3600`, `86400`,
        `259200`, `604800`.
    name : `str`
        The channel's name.
    open : `bool`
        Whether the thread channel is open.
    owner_id : `int`
        The channel's creator's identifier. Defaults to `0`.
    parent_id : `int`
        The channel's parent's identifier.
    slowmode : `int`
        The amount of time in seconds what a user needs to wait between it's each message. Bots and user accounts with
        `manage_messages`, `manage_channels` permissions are unaffected.
    thread_users : `None`, `dict` of (`int`, ``ClientUserBase``) items
        The users inside of the thread if any.
    
    Class Attributes
    ----------------
    order_group: `int` = `0`
        The channel's order group used when sorting channels.
    """
    __slots__ = ()



__all__ = ('Connection', )

from scarletio import include

from ...bases import DiscordEntity
from ...precreate_helpers import process_precreate_parameters_and_raise_extra

from .fields import (
    parse_friend_sync, parse_id, parse_integrations, parse_metadata_visibility, parse_name, parse_revoked,
    parse_show_activity, parse_two_way_link, parse_type, parse_verified, parse_visibility, put_friend_sync,
    put_id, put_integrations, put_metadata_visibility, put_name, put_revoked,
    put_show_activity, put_two_way_link, put_type, put_verified, put_visibility,
    validate_friend_sync, validate_id, validate_integrations, validate_metadata_visibility, validate_name,
    validate_revoked, validate_show_activity, validate_two_way_link, validate_type, validate_verified,
    validate_visibility
)
from .preinstanced import ConnectionType, ConnectionVisibility


Integration = include('Integration')

PRECREATE_FIELDS = {
    'connection_type': ('type', validate_type),
    'friend_sync': ('friend_sync', validate_friend_sync),
    'integrations': ('integrations', validate_integrations),
    'metadata_visibility': ('metadata_visibility', validate_metadata_visibility),
    'name': ('name', validate_name),
    'revoked': ('revoked', validate_revoked),
    'show_activity': ('show_activity', validate_show_activity),
    'two_way_link': ('two_way_link', validate_two_way_link),
    'verified': ('verified', validate_verified),
    'visibility': ('visibility', validate_visibility),
}


class Connection(DiscordEntity):
    """
    A connection object that a user is attached to.
    
    Attributes
    ----------
    friend_sync : `bool`
        Whether the user has friend sync enabled for the connection.
    
    id : `int`
        The unique identifier value of the connection.
    
    integrations : `None` or `tuple` of ``Integration``
        A guild's integrations which are attached to the connection.
    
    metadata_visibility : ``ConnectionVisibility``
        For who is the connection metadata visible for.
    
    name : `str`
        The username of the connected account.
    
    revoked : `bool`
        Whether the connection is revoked.
    
    show_activity : `bool`
        Whether activity related to this connection will be shown in presence updates.
    
    two_way_link : `bool`
         Whether this connection has a corresponding third party OAuth2 token.
    
    type : ``ConnectionType``
        The service of the connection.
    
    verified : `bool`
        Whether the connection is verified.
    
    visibility : ``ConnectionVisibility``
        For who is the connection visible for.
    """
    __slots__ = (
        'friend_sync', 'integrations', 'metadata_visibility', 'name', 'revoked', 'show_activity', 'two_way_link',
        'type', 'verified', 'visibility'
    )
    
    
    def __new__(
        cls,
        *,
        connection_type = ...,
        friend_sync = ...,
        integrations = ...,
        metadata_visibility = ...,
        name = ...,
        revoked = ...,
        show_activity = ...,
        two_way_link = ...,
        verified = ...,
        visibility = ...,
    ):
        """
        Creates a new connection from the given parameters.
        
        Parameters
        ----------
        connection_type : ``ConnectionType``, `str`, Optional (Keyword only)
            The service of the connection.
        
        friend_sync : `bool`, Optional (Keyword only)
            Whether the user has friend sync enabled for the connection.
        
        integrations : `None` or `iterable` of ``Integration``, Optional (Keyword only)
            A guild's integrations which are attached to the connection.
        
        metadata_visibility : ``ConnectionVisibility``, `int`, Optional (Keyword only)
            For who is the connection metadata visible for.
        
        name : `str`, Optional (Keyword only)
            The username of the connected account.
        
        revoked : `bool`, Optional (Keyword only)
            Whether the connection is revoked.
        
        show_activity : `bool`, Optional (Keyword only)
            Whether activity related to this connection will be shown in presence updates.
        
        two_way_link : `bool`, Optional (Keyword only)
             Whether this connection has a corresponding third party OAuth2 token.
        
        verified : `bool`, Optional (Keyword only)
            Whether the connection is verified.
        
        visibility : ``ConnectionVisibility``, `int`, Optional (Keyword only)
            For who is the connection visible for.
        
        Raises
        ------
        TypeError
            - If a field's type is unacceptable.
        ValueError
            - If a a field's value is unacceptable.
        """
        # connection_type
        if connection_type is ...:
            connection_type = ConnectionType.none
        else:
            connection_type = validate_type(connection_type)
        
        # friend_sync
        if friend_sync is ...:
            friend_sync = False
        else:
            friend_sync = validate_friend_sync(friend_sync)
        
        # integrations
        if integrations is ...:
            integrations = None
        else:
            integrations = validate_integrations(integrations)
        
        # metadata_visibility
        if metadata_visibility is ...:
            metadata_visibility = ConnectionVisibility.user_only
        else:
            metadata_visibility = validate_metadata_visibility(metadata_visibility)
        
        # name
        if name is ...:
            name = ''
        else:
            name = validate_name(name)
        
        # revoked
        if revoked is ...:
            revoked = False
        else:
            revoked = validate_revoked(revoked)
        
        # show_activity
        if show_activity is ...:
            show_activity = False
        else:
            show_activity = validate_show_activity(show_activity)
        
        # two_way_link
        if two_way_link is ...:
            two_way_link = False
        else:
            two_way_link = validate_two_way_link(two_way_link)
        
        # verified
        if verified is ...:
            verified = False
        else:
            verified = validate_verified(verified)
        
        # visibility
        if visibility is ...:
            visibility = ConnectionVisibility.user_only
        else:
            visibility = validate_visibility(visibility)
        
        # Construct
        self = object.__new__(cls)
        
        self.id = 0
        self.friend_sync = friend_sync
        self.integrations = integrations
        self.metadata_visibility = metadata_visibility
        self.name = name
        self.revoked = revoked
        self.show_activity = show_activity
        self.two_way_link = two_way_link
        self.type = connection_type
        self.verified = verified
        self.visibility = visibility
        
        return self
    
    
    @classmethod
    def precreate(cls, connection_id, **keyword_parameters):
        """
        Creates a new connection with the given predefined fields.
        
        > Since connections are not cached globally, this method is only be used for testing.
        
        Parameters
        ----------
        connection_id : `int`
            The connection's identifier.
        **keyword_parameters : Keyword Parameters
            The attributes to set.
        
        Other Parameters
        ----------------
        connection_type : ``ConnectionType``, `str`, Optional (Keyword only)
            The service of the connection.
        
        friend_sync : `bool`, Optional (Keyword only)
            Whether the user has friend sync enabled for the connection.
        
        integrations : `None` or `iterable` of ``Integration``, Optional (Keyword only)
            A guild's integrations which are attached to the connection.
        
        metadata_visibility : ``ConnectionVisibility``, `int`, Optional (Keyword only)
            For who is the connection metadata visible for.
        
        name : `str`, Optional (Keyword only)
            The username of the connected account.
        
        revoked : `bool`, Optional (Keyword only)
            Whether the connection is revoked.
        
        show_activity : `bool`, Optional (Keyword only)
            Whether activity related to this connection will be shown in presence updates.
        
        two_way_link : `bool`, Optional (Keyword only)
             Whether this connection has a corresponding third party OAuth2 token.
        
        verified : `bool`, Optional (Keyword only)
            Whether the connection is verified.
        
        visibility : ``ConnectionVisibility``, `int`, Optional (Keyword only)
            For who is the connection visible for.
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - If `connection_id` is not `int`.
            - If a field's type is unacceptable.
            - Extra or unused parameters
        ValueError
            - If a a field's value is unacceptable.
        """
        connection_id = validate_id(connection_id)
        
        if keyword_parameters:
            processed = process_precreate_parameters_and_raise_extra(keyword_parameters, PRECREATE_FIELDS)
        else:
            processed = None
        
        self = cls._create_empty(connection_id)
        
        if (processed is not None):
            for name, value in processed:
                setattr(self, name, value)
        
        return self
    
    
    @classmethod
    def _create_empty(cls, connection_id):
        """
        Creates a new connection with it's default attributes set.
        
        Parameters
        ----------
        connection_id : `int`
            The connection's identifier.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        
        self.id = connection_id
        self.friend_sync = False
        self.integrations = None
        self.metadata_visibility = ConnectionVisibility.user_only
        self.name = ''
        self.revoked = False
        self.show_activity = False
        self.two_way_link = False
        self.type = ConnectionType.none
        self.verified = False
        self.visibility = ConnectionVisibility.user_only
        
        return self
    
    
    def __repr__(self):
        """Returns the connection's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        connection_id = self.id
        if connection_id:
            repr_parts.append(' id = ')
            repr_parts.append(repr(connection_id))
        
        else:
            repr_parts.append(' (partial)')
        
        connection_type = self.type
        if (connection_type is not ConnectionType.none):
            repr_parts.append(', type = ')
            repr_parts.append(connection_type.name)
            repr_parts.append(' ~ ')
            repr_parts.append(repr(connection_type.value))
        
        name = self.name
        if name:
            repr_parts.append(', name = ')
            repr_parts.append(repr(name))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    def __hash__(self):
        """Returns the connection's hash."""
        connection_id = self.id
        if connection_id:
            return connection_id
        
        return self._get_hash_partial()
    
    
    def _get_hash_partial(self):
        """
        Returns a partial connection's hash value.
        
        Returns
        -------
        hash_value : `int`
        """
        hash_value = 0
        
        # friend_sync
        hash_value ^= self.friend_sync
        
        # id | Internal -> skip
        
        # integrations
        integrations = self.integrations
        if (integrations is not None):
            hash_value ^= len(integrations) << 1
            for integration in integrations:
                hash_value ^= hash(integration)
        
        # metadata_visibility
        hash_value ^= hash(self.metadata_visibility) << 5
        
        # name
        hash_value ^= hash(self.name)
        
        # revoked
        hash_value ^= self.revoked << 9
        
        # show_activity
        hash_value ^= self.show_activity << 10
        
        # two_way_link
        hash_value ^= self.two_way_link << 11
        
        # type
        hash_value ^= hash(self.type)
        
        # verified
        hash_value ^= self.verified << 12
        
        # visibility
        hash_value ^= hash(self.visibility) << 13
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two connections are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal_same_type(other)
    
    
    def __ne__(self, other):
        """Returns whether the two connections are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return not self._is_equal_same_type(other)
    
    
    def _is_equal_same_type(self, other):
        """
        Helper method for ``.__eq__``
        
        Parameters
        ----------
        other : `instance<type<<self>>`
            The other instance. Must be from the same type.
        
        Returns
        -------
        is_equal : `bool`
        """
        # id
        self_id = self.id
        other_id = other.id
        if self_id and other_id:
            if self_id == other_id:
                return True
            
            return False
        
        # friend_sync
        if self.friend_sync != other.friend_sync:
            return False
        
        # integrations
        if self.integrations != other.integrations:
            return False
        
        # metadata_visibility
        if self.metadata_visibility is not other.metadata_visibility:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        # revoked
        if self.revoked != other.revoked:
            return False
        
        # show_activity
        if self.show_activity != other.show_activity:
            return False
        
        # two_way_link
        if self.two_way_link != other.two_way_link:
            return False
        
        # type
        if self.type is not other.type:
            return False
        
        # verified
        if self.verified != other.verified:
            return False
        
        # visibility
        if self.visibility is not other.visibility:
            return False
        
        return True
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a connection object from received connection data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`)
            Received connection data.
        """
        self = object.__new__(cls)
        
        self.id = parse_id(data)
        self.friend_sync = parse_friend_sync(data)
        self.integrations = parse_integrations(data)
        self.metadata_visibility = parse_metadata_visibility(data)
        self.name = parse_name(data)
        self.revoked = parse_revoked(data)
        self.show_activity = parse_show_activity(data)
        self.two_way_link = parse_two_way_link(data)
        self.type = parse_type(data)
        self.verified = parse_verified(data)
        self.visibility = parse_visibility(data)
        
        return self
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Converts the connection to json serializable representation dictionary.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether we want to include identifiers as well.
        
        Returns
        -------
        data : `dict` of (`str`, `str`) items
        """
        data = {}
        
        if include_internals:
            put_id(self.id, data, defaults)
        
        put_friend_sync(self.friend_sync, data, defaults)
        put_integrations(self.integrations, data, defaults, include_internals = include_internals)
        put_metadata_visibility(self.metadata_visibility, data, defaults)
        put_name(self.name, data, defaults)
        put_revoked(self.revoked, data, defaults)
        put_show_activity(self.show_activity, data, defaults)
        put_two_way_link(self.two_way_link, data, defaults)
        put_type(self.type, data, defaults)
        put_verified(self.verified, data, defaults)
        put_visibility(self.visibility, data, defaults)
        
        return data
    
    
    def iter_integrations(self):
        """
        Iterates over the integrations of the connection.
        
        This method is an iterable generator.
        
        Yields
        ------
        integration : ``Integration``
        """
        integrations = self.integrations
        if (integrations is not None):
            yield from integrations
    
    
    def copy(self):
        """
        Copies the connection returning a new partial one.
        
        Returns
        -------
        new : `instance<cls>`
        """
        # Construct
        new = object.__new__(type(self))
        new.id = 0
        new.friend_sync = self.friend_sync
        integrations = self.integrations
        if (integrations is not None):
            integrations = (*integrations,)
        new.integrations = integrations
        new.metadata_visibility = self.metadata_visibility
        new.name = self.name
        new.revoked = self.revoked
        new.show_activity = self.show_activity
        new.two_way_link = self.two_way_link
        new.type = self.type
        new.verified = self.verified
        new.visibility = self.visibility
        return new

    
    def copy_with(
        self,
        *,
        connection_type = ...,
        friend_sync = ...,
        integrations = ...,
        metadata_visibility = ...,
        name = ...,
        revoked = ...,
        show_activity = ...,
        two_way_link = ...,
        verified = ...,
        visibility = ...,):
        """
        Copies the connection with the given fields returning a new partial one.
        
        Parameters
        ----------
        connection_type : ``ConnectionType``, `str`, Optional (Keyword only)
            The service of the connection.
        
        friend_sync : `bool`, Optional (Keyword only)
            Whether the user has friend sync enabled for the connection.
        
        integrations : `None` or `iterable` of ``Integration``, Optional (Keyword only)
            A guild's integrations which are attached to the connection.
        
        metadata_visibility : ``ConnectionVisibility``, `int`, Optional (Keyword only)
            For who is the connection metadata visible for.
        
        name : `str`, Optional (Keyword only)
            The username of the connected account.
        
        revoked : `bool`, Optional (Keyword only)
            Whether the connection is revoked.
        
        show_activity : `bool`, Optional (Keyword only)
            Whether activity related to this connection will be shown in presence updates.
        
        two_way_link : `bool`, Optional (Keyword only)
             Whether this connection has a corresponding third party OAuth2 token.
        
        verified : `bool`, Optional (Keyword only)
            Whether the connection is verified.
        
        visibility : ``ConnectionVisibility``, `int`, Optional (Keyword only)
            For who is the connection visible for.
        
        Raises
        ------
        TypeError
            - If a field's type is unacceptable.
        ValueError
            - If a a field's value is unacceptable.
        
        Returns
        -------
        new : `instance<cls>`
        """
        # connection_type
        if connection_type is ...:
            connection_type = self.type
        else:
            connection_type = validate_type(connection_type)
        
        # friend_sync
        if friend_sync is ...:
            friend_sync = self.friend_sync
        else:
            friend_sync = validate_friend_sync(friend_sync)
        
        # integrations
        if integrations is ...:
            integrations = self.integrations
            if (integrations is not None):
                integrations = (*integrations,)
        else:
            integrations = validate_integrations(integrations)
        
        # metadata_visibility
        if metadata_visibility is ...:
            metadata_visibility = self.metadata_visibility
        else:
            metadata_visibility = validate_metadata_visibility(metadata_visibility)
        
        # name
        if name is ...:
            name = self.name
        else:
            name = validate_name(name)
        
        # revoked
        if revoked is ...:
            revoked = self.revoked
        else:
            revoked = validate_revoked(revoked)
        
        # show_activity
        if show_activity is ...:
            show_activity = self.show_activity
        else:
            show_activity = validate_show_activity(show_activity)
        
        # two_way_link
        if two_way_link is ...:
            two_way_link = self.two_way_link
        else:
            two_way_link = validate_two_way_link(two_way_link)
        
        # verified
        if verified is ...:
            verified = self.verified
        else:
            verified = validate_verified(verified)
        
        # visibility
        if visibility is ...:
            visibility = self.visibility
        else:
            visibility = validate_visibility(visibility)
        
        # Construct
        new = object.__new__(type(self))
        new.id = 0
        new.friend_sync = friend_sync
        integrations = integrations
        new.integrations = integrations
        new.metadata_visibility = metadata_visibility
        new.name = name
        new.revoked = revoked
        new.show_activity = show_activity
        new.two_way_link = two_way_link
        new.type = connection_type
        new.verified = verified
        new.visibility = visibility
        return new



__all__ = ()

from ...channel import PermissionOverwriteTargetType
from ...channel.permission_overwrite.fields import validate_target_id, validate_target_type
from ...role.role.fields import validate_name

from ..audit_log_entry_detail_conversion import AuditLogEntryDetailConversion, AuditLogEntryDetailConversionGroup
from ..conversion_helpers.converters import (
    value_deserializer_id, value_deserializer_name, value_serializer_id, value_serializer_name
)


# ---- role_name ----

ROLE_NAME_CONVERSION = AuditLogEntryDetailConversion(
    'role_name',
    'role_name',
    value_deserializer = value_deserializer_name,
    value_serializer = value_serializer_name,
    value_validator = validate_name,
)


# ---- target_id ----

TARGET_ID_CONVERSION = AuditLogEntryDetailConversion(
    'id',
    'target_id',
    value_deserializer = value_deserializer_id,
    value_serializer = value_serializer_id,
    value_validator = validate_target_id,
)


# ---- target_type ----

TARGET_TYPE_CONVERSION = AuditLogEntryDetailConversion(
    'type',
    'target_type',
    value_validator = validate_target_type,
)

@TARGET_TYPE_CONVERSION.set_value_deserializer
def target_type_value_deserializer(value):
    if (value is not None) and (PermissionOverwriteTargetType.VALUE_TYPE is int):
        value = int(value)
    
    return PermissionOverwriteTargetType(value)


@TARGET_TYPE_CONVERSION.set_value_serializer
def target_type_value_serializer(value):
    return str(value.value)


# ---- Construct ----

CHANNEL_PERMISSION_OVERWRITE_CONVERSIONS = AuditLogEntryDetailConversionGroup(
    ROLE_NAME_CONVERSION,
    TARGET_ID_CONVERSION,
    TARGET_TYPE_CONVERSION,
)



__all__ = ()

from ...field_parsers import entity_id_parser_factory
from ...field_putters import entity_id_putter_factory
from ...field_validators import entity_id_validator_factory
from ...user import ClientUserBase

from ..guild import Guild

# guild_id

parse_guild_id = entity_id_parser_factory('guild_id')
put_guild_id = entity_id_putter_factory('guild_id')
validate_guild_id = entity_id_validator_factory('guild_id', Guild)

# user_id

parse_user_id = entity_id_parser_factory('user_id')
put_user_id = entity_id_putter_factory('user_id')
validate_user_id = entity_id_validator_factory('user_id', ClientUserBase)



__all__ = ()

from .fields import (
    put_emoji_discovery, put_keywords, put_primary_category, validate_emoji_discovery,
    validate_keywords, validate_primary_category
)


GUILD_DISCOVERY_FIELD_CONVERTERS = {
    'emoji_discovery': (validate_emoji_discovery, put_emoji_discovery),
    'keywords': (validate_keywords, put_keywords),
    'primary_category': (validate_primary_category, put_primary_category),
}



__all__ = ()

from ...field_parsers import force_string_parser_factory
from ...field_putters import url_optional_putter_factory
from ...field_validators import (
    int_conditional_validator_factory, nullable_string_validator_factory, url_required_validator_factory
)
from ...user.user.fields import (
    parse_discriminator, parse_id, parse_name, parse_status, put_discriminator, put_id, put_name,
    put_status, validate_discriminator, validate_name, validate_status
)


# activity_name

def parse_activity_name(data):
    """
    Parses out the `activity_name` field from the given data.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        Guild widget user data.
    
    Returns
    -------
    activity_name : `None`, `str`
    """
    activity_data = data.get('game', None)
    if activity_data is None:
        activity_name = None
    else:
        activity_name = activity_data.get('name', None)
        if (activity_name is not None) and (not activity_name):
            activity_name = None
    
    return activity_name


def put_activity_name(activity_name, data, defaults):
    """
    Puts the given activity name into a json serializable dictionary.
    
    Parameters
    ----------
    activity_name : `str`
        Activity's name.
    data : `dict` of (`str`, `object`) items
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    if (activity_name is not None) or defaults:
        if activity_name is None:
            activity_name = ''
        data['game'] = {'name': activity_name}
    
    return data


validate_activity_name = nullable_string_validator_factory('activity_name', 0, 1024)

# id

validate_id = int_conditional_validator_factory(
    'id',
    0,
    lambda id : id >= 0 and id <= 99,
    '>= 0 and <= 99',
)

# url

parse_avatar_url = force_string_parser_factory('avatar_url')
put_avatar_url = url_optional_putter_factory('avatar_url')
validate_avatar_url = url_required_validator_factory('avatar_url')



__all__ = ('OnboardingPromptType',)

from scarletio import export

from ...bases import Preinstance as P, PreinstancedBase


@export
class OnboardingPromptType(PreinstancedBase, value_type = int):
    """
    The type of an onboarding prompt.
    
    Attributes
    ----------
    name : `str`
        The name of the onboarding prompt type.
    
    value : `int`
        The identifier value the onboarding prompt type.
    
    Type Attributes
    ---------------
    Every predefined onboarding prompt type can be accessed as type attribute as well:
    
    +-----------------------------------+-----------------------------------+-------+
    | Type attribute name               | Name                              | Value |
    +===================================+===================================+=======+
    | multiple_choice                   | multiple choice                   | 0     |
    +-----------------------------------+-----------------------------------+-------+
    | dropdown                          | dropdown                          | 1     |
    +-----------------------------------+-----------------------------------+-------+
    """
    __slots__ = ()
    
    multiple_choice = P(0, 'multiple choice')
    dropdown = P(1, 'dropdown')



__all__ = ('PollAnswer',)

from ...bases import DiscordEntity
from ...precreate_helpers import process_precreate_parameters_and_raise_extra
from ...utils import sanitize_mentions

from .fields import (
    parse_emoji, parse_id, parse_text, put_emoji, put_id, put_text, validate_emoji, validate_id,
    validate_text
)


PRECREATE_FIELDS = {
    'emoji': ('emoji', validate_emoji),
    'text': ('text', validate_text),
}


class PollAnswer(DiscordEntity):
    """
    Represents a poll's answer.
    
    Attributes
    ----------
    emoji : `None`, ``Emoji``
        The poll answer's emoji.
    id : `int`
        The answer's identifier.
    text : `None`, `str`
        The poll answer's text.
    """
    __slots__ = ('emoji', 'text',)
    
    def __new__(cls, *, emoji = ..., text = ...):
        """
        Creates a new poll answer with the given parameters.
        
        Parameters
        ----------
        emoji : `None`, ``Emoji``, Optional (Keyword only)
            The answer's emoji.
        text : `None`, `str`, Optional (Keyword only)
            The answer's text.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # emoji
        if emoji is ...:
            emoji = None
        else:
            emoji = validate_emoji(emoji)
        
        # text
        if text is ...:
            text = None
        else:
            text = validate_text(text)
        
        self = object.__new__(cls)
        self.emoji = emoji
        self.id = 0
        self.text = text
        return self
    
    
    def __len__(self):
        """Returns the poll answer's length."""
        length = 0
        
        # emoji
        # does not count
        
        # id
        # does not count
        
        # text
        text = self.text
        if (text is not None):
            length += len(text)
        
        return length
    
    
    def __bool__(self):
        """Returns whether the poll answer has any fields set."""
        # emoji
        if self.emoji is not None:
            return True
        
        # id
        # does not count
        
        # text
        if self.text is not None:
            return True
        
        return False
    
    
    def __repr__(self):
        repr_parts = ['<', type(self).__name__]
        
        # id
        answer_id = self.id
        if answer_id:
            repr_parts.append(' id = ')
            repr_parts.append(repr(answer_id))
            
            field_added = True
        else:
            field_added = False
        
        # emoji
        emoji = self.emoji
        if (emoji is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' emoji = ')
            repr_parts.append(repr(emoji))
        
        # text
        text = self.text
        if (text is not None):
            if field_added:
                repr_parts.append(',')
            
            repr_parts.append(' text = ')
            repr_parts.append(repr(text))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the poll answer's hash value."""
        hash_value = 0
        
        # emoji
        emoji = self.emoji
        if (emoji is not None):
            hash_value ^= hash(emoji)
        
        # id
        hash_value ^= self.id
        
        # text
        text = self.text
        if (text is not None):
            hash_value ^= hash(text)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two poll answers are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # id
        self_id = self.id
        other_id = other.id
        if (self_id and other_id) and (self_id != other_id):
            return False
        
        # emoji
        if self.emoji is not other.emoji:
            return False
        
        # text
        if self.text != other.text:
            return False
        
        return True
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new poll answer instance from the given json data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Data to create poll answer from.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.emoji = parse_emoji(data)
        self.id = parse_id(data)
        self.text = parse_text(data)
        return self
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Returns the poll answer as a json serializable representation.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields with their default values should be included as well.
        
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        
        put_emoji(self.emoji, data, defaults)
        put_text(self.text, data, defaults)
        
        if include_internals:
            put_id(self.id, data, defaults)
        
        return data
    
    
    def clean_copy(self, guild = None):
        """
        Creates a clean copy of the poll answer by removing the mentions in it's contents.
        
        Parameters
        ----------
        guild : `None`, ``Guild`` = `None`, Optional
            The respective guild as a context to look up guild specific names of entities.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.emoji = self.emoji
        new.id = 0
        new.text = sanitize_mentions(self.text, guild)
        return new
    
    
    def copy(self):
        """
        Copies the poll answer.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.emoji = self.emoji
        new.id = 0
        new.text = self.text
        return new
    
    
    def copy_with(self, *, emoji = ..., text = ...):
        """
        Copies the poll answer with the given parameters.
        
        Parameters
        ----------
        emoji : `None`, ``Emoji``, Optional (Keyword only)
            The answer's emoji.
        text : `None`, `str`, Optional (Keyword only)
            The answer's text.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # emoji
        if emoji is ...:
            emoji = self.emoji
        else:
            emoji = validate_emoji(emoji)
        
        # text
        if text is ...:
            text = self.text
        else:
            text = validate_text(text)
        
        new = object.__new__(type(self))
        new.emoji = emoji
        new.id = 0
        new.text = text
        return new
    
    
    @classmethod
    def precreate(
        cls,
        answer_id,
        **keyword_parameters,
    ):
        """
        Precreates an answer. Since answers are not cached, this method just a ``.__new__`` alternative.
        
        Parameters
        ----------
        answer_id : `int`
            The answer's identifier.
        
        **keyword_parameters : Keyword parameters
            Additional parameters defining how the answer's fields should be set.
        
        Other Parameters
        ----------------
        emoji : `None`, ``Emoji``, Optional (Keyword only)
            The answer's emoji.
        text : `None`, `str`, Optional (Keyword only)
            The answer's text.
        
        Returns
        -------
        self : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        answer_id = validate_id(answer_id)
        
        if keyword_parameters:
            processed = process_precreate_parameters_and_raise_extra(keyword_parameters, PRECREATE_FIELDS)
        
        else:
            processed = None
        
        # Construct
        
        self = object.__new__(cls)
        self.emoji = None
        self.id = answer_id
        self.text = None
        
        if (processed is not None):
            for item in processed:
                setattr(self, *item)
        
        return self

    
    @property
    def contents(self):
        """
        Returns the contents of the poll answer.
        
        Returns
        -------
        contents : `list<str>`
        """
        return [*self.iter_contents()]
    
    
    def iter_contents(self):
        """
        Iterates over the contents of the poll answer.
        
        This method is an iterable generator.
        
        Yields
        ------
        content : `str`
        """
        text = self.text
        if (text is not None):
            yield text



from .application_command import *
from .constants import *
from .fields import *
from .preinstanced import *


__all__ = (
    *application_command.__all__,
    *constants.__all__,
    *fields.__all__,
    *preinstanced.__all__,
)



__all__ = ()

from scarletio import Compound, Theory

from ...application import Application
from ...bases import maybe_snowflake
from ...http import DiscordApiClient
from ...user import ClientUserBase, HypesquadHouse, RelationshipType
from ...utils import Relationship
from ..request_helpers import get_user_and_id, get_user_id


class ClientCompoundLockedEndpoints(Compound):
    
    api : DiscordApiClient
    
    @Theory
    async def user_get(self, user, *, force_update = False): ...
    
    
    async def relationship_delete(self, relationship):
        """
        Deletes the given relationship.
        
        This method is a coroutine.
        
        Parameters
        ----------
        relationship : ``Relationship``, ``ClientUserBase``, `int`
            The relationship to delete. Also can be given the respective user with who the client has the relationship
            with.

        Raises
        ------
        TypeError
            If `relationship` was not given neither as ``Relationship``, ``ClientUserBase`` not as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        This endpoint is available only for user accounts.
        """
        if isinstance(relationship, Relationship):
            user_id = relationship.user.id
        elif isinstance(relationship, ClientUserBase):
            user_id = relationship.id
        else:
            user_id = maybe_snowflake(relationship)
            if user_id is None:
                raise TypeError(
                    f'`relationship` can be `{Relationship.__name__}`, `{ClientUserBase.__name__}`, `int`, got '
                    f'{relationship.__class__.__name__}; {relationship!r}.'
                )
        
        await self.api.relationship_delete(user_id)
    
    
    async def relationship_create(self, user, relationship_type = None):
        """
        Creates a relationship with the given user.
        
        This method is a coroutine.
        
        Parameters
        ----------
        user : ```ClientUserBase``, `int`
            The user with who the relationship will be created.
        relationship_type : `None`, ``RelationshipType``, `int` = `None`, Optional
            The type of the relationship. Defaults to `None`.
        
        Raises
        ------
        TypeError
            - If `user` is not given neither as ``ClientUserBase``, `int`.
            - If `relationship_type` was not given neither as `None`, ``RelationshipType`` neither as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        This endpoint is available only for user accounts.
        """
        user_id = get_user_id(user)
        
        if relationship_type is None:
            relationship_type_value = None
        elif isinstance(relationship_type, RelationshipType):
            relationship_type_value = relationship_type.value
        elif isinstance(relationship_type, int):
            relationship_type_value = relationship_type
        else:
            raise TypeError(
                f'`relationship_type` can be `None`, `{RelationshipType.__name__}`, `int`'
                f', got {relationship_type.__class__.__name__}; {relationship_type!r}.'
            )
        
        data = {}
        if (relationship_type_value is not None):
            data['type'] = relationship_type_value
        
        await self.api.relationship_create(user_id, data)
    
    
    async def relationship_friend_request(self, user):
        """
        Sends a friend request to the given user.
        
        This method is a coroutine.
        
        Parameters
        ----------
        user : ```ClientUserBase``, `int`
            The user, who will receive the friend request.
        
        Raises
        ------
        TypeError
            If `user` was not given neither as ```ClientUserBase``, `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        This endpoint is available only for user accounts.
        """
        user, user_id = get_user_and_id(user)
        if (user is None):
            user = await self.user_get(user_id)
        
        data = {
            'username': user.name,
            'discriminator': str(user.discriminator)
        }
        
        await self.api.relationship_friend_request(data)


    async def hypesquad_house_change(self, house):
        """
        Changes the client's hypesquad house.
        
        This method is a coroutine.
        
        Parameters
        ----------
        house : `int`, ``HypesquadHouse``
            The hypesquad house to join.
        
        Raises
        ------
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Raises
        ------
        TypeError
            `house` was not given as `int`  neither as ``HypesquadHouse``.
        
        Notes
        -----
        User account only.
        """
        if isinstance(house, HypesquadHouse):
            house_id = house.value
        elif isinstance(house, int):
            house_id = house
        else:
            raise TypeError(
                f'`house` can be `int`, `{HypesquadHouse.__name__}`, got {house.__class__.__name__}; {house!r}.'
            )
        
        await self.api.hypesquad_house_change({'house_id': house_id})
    
    
    async def hypesquad_house_leave(self):
        """
        Leaves the client from it's current hypesquad house.
        
        This method is a coroutine.
        
        Raises
        ------
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        User account only.
        """
        await self.api.hypesquad_house_leave()


    async def application_get(self, application):
        """
        Requests a specific application by it's id.
        
        This method is a coroutine.
        
        Parameters
        ----------
        application : ``Application``, `int`
            The application or it's identifier to request.
        
        Returns
        -------
        application : ``Application``
        
        Raises
        ------
        TypeError
            If `application` was not given neither as ``Application`` nor as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        This endpoint does not support bot accounts.
        """
        if isinstance(application, Application):
            application_id = application.id
        else:
            application_id = maybe_snowflake(application)
            if application_id is None:
                raise TypeError(
                    f'`application` can be `{Application.__name__}`, `int`, got '
                    f'{application.__class__.__name__}; {application!r}.'
                )
        
        application_data = await self.api.application_get(application_id)
        return Application.from_data_own(application_data)



__all__ = ('EnvGetter',)

from warnings import warn

from scarletio import RichAttributeErrorBaseType

from .getters import (
    RETURN_TYPE_EXCEPTION, RETURN_TYPE_VALUE, RETURN_TYPE_WARNING, _get_env, _handle_get_env_generator,
    _process_bool_env, _process_int_env, _process_str_env
)


class EnvGetter(RichAttributeErrorBaseType):
    """
    Env variable getter that can be used to aggregate error messages.
    
    Attributes
    ----------
    _captured : `None, `list<(int, str)>`
        Captured errors.
    _entered : `int`
        Whether the getter is entered as a context manager. If not, it will propagate the errors instantly.
    
    Examples
    --------
    ```py
    with EnvGetter() as env:
        KOISHI_TOKEN = env.get_str('KOISHI_TOKEN')
        SATORI_TOKEN = env.get_str('SATORI_TOKEN')
    """
    __slots__ = ('_captured', '_entered')
    
    def __new__(cls):
        """
        Creates a new env getter instance.
        
        Env getter can be used to aggregate error messages.
        """
        self = object.__new__(cls)
        self._captured = None
        self._entered = 0
        return self
    
    
    def __repr__(self):
        """Returns the env getter's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        if self._entered:
            repr_parts.append(' entered')
            
            field_added = True
        else:
            field_added = False
        
        captured = self._captured
        if (captured is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' captured = ')
            repr_parts.append(repr(captured))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __enter__(self):
        """Enters the context manager."""
        self._entered += 1
        return self
    
    
    def __exit__(self, exception_type, exception_value, exception_traceback):
        """
        Exits the context manager.
        
        Parameters
        ----------
        exception_type : `None | type<BaseException>`
            The occurred exception's type if any.
        
        exception_value : `None | BaseException`
            The occurred exception if any.
        
        exception_traceback : `None | TracebackType`
            the exception's traceback if any.
        
        Returns
        -------
        captured : `bool`
            Whether the exception was captured.
        """
        entered = self._entered - 1
        self._entered = entered
        
        if entered:
            return False
        
        error_type, error_message = self._get_aggregated_state()
        self._captured = None
        
        if error_type == RETURN_TYPE_WARNING:
            warn(error_message, stacklevel = 2)
        elif error_type == RETURN_TYPE_EXCEPTION:
            raise RuntimeError(error_message)
        
        return False
    
    
    def get_bool(self, name, default = False, *, raise_if_missing_or_empty = False, warn_if_empty = True):
        """
        Gets the given environmental variable.
        
        If the environmental variable is not present or not present as a boolean's representation returns `default`
        instead.
        
        Parameters
        ----------
        name : `str`
            The name of an environmental variable.
        default : `bool` = `False`, Optional
            The default value of the respective variable.
        raise_if_missing_or_empty : `bool` = `False`, Optional (Keyword only)
            Whether exception should be thrown if the environmental variable is missing or empty.
            Has priority over `default` and `warn_if_empty`.
        warn_if_empty : `bool` = `True`, Optional (Keyword only)
            Whether warning should be dropped if empty environmental variable is received.
        
        Returns
        -------
        variable : `bool`
        """
        return self._handle(
            _get_env(name, default, 'bool', _process_bool_env, raise_if_missing_or_empty, warn_if_empty), default
        )

    
    def get_str(self, name, default = False, *, raise_if_missing_or_empty = False, warn_if_empty = True):
        """
        Gets the given environmental variable.
        
        If the environmental variable is not present or present as an empty string returns `default` instead.
        
        Parameters
        ----------
        name : `str`
            The name of an environmental variable.
        default : `None` | `str` = `None`, Optional
            The default value of the respective variable. Defaults to `None`.
        raise_if_missing_or_empty : `bool` = `False`, Optional (Keyword only)
            Whether exception should be thrown if the environmental variable is missing or empty.
            Has priority over `default` and `warn_if_empty`.
        warn_if_empty : `bool` = `True`, Optional (Keyword only)
            Whether warning should be dropped if empty environmental variable is received.
            
        Returns
        -------
        variable : `bool`
        """
        return self._handle(
            _get_env(name, default, 'str', _process_str_env, raise_if_missing_or_empty, warn_if_empty), default
        )
    
    
    def get_int(self, name, default = 0, *, raise_if_missing_or_empty = False, warn_if_empty = True):
        """
        Gets the given environmental variable.
        
        If the environmental variable is not present or present as non `int`, will return `default` instead.
        
        Parameters
        ----------
        name : `str`
            The name of an environmental variable.
        default : `int` = `0`, Optional
            The default value of the respective variable.
        raise_if_missing_or_empty : `bool` = `False`, Optional (Keyword only)
            Whether exception should be thrown if the environmental variable is missing or empty.
            Has priority over `default` and `warn_if_empty`.
        warn_if_empty : `bool` = `True`, Optional (Keyword only)
            Whether warning should be dropped if empty environmental variable is received.
        
        Returns
        -------
        variable : `int`
        """
        return self._handle(
            _get_env(name, default, 'int', _process_int_env, raise_if_missing_or_empty, warn_if_empty), default
        )
    
    
    def _handle(self, generator, default):
        """
        Handles a ``_get_env`` generator.
        
        Parameters
        ----------
        generator : `iterable<(int, object)>`
            Generator to iterate over.
        default : `object`
            Default to return in case we are capturing the errors & warnings.
        
        Returns
        -------
        value : `object`
        
        Raises
        ------
        RuntimeError
        """
        if not self._entered:
            return _handle_get_env_generator(generator)
        
        return self._handle_as_entered(generator, default)
    
    
    def _handle_as_entered(self, generator, default):
        """
        Handles a ``_get_env`` generator as capturing its errors & warnings.
        
        Parameters
        ----------
        generator : `iterable<(int, object)>`
            Generator to iterate over.
        default : `object`
            Default to return in case we are capturing the errors & warnings.
        
        Returns
        -------
        value : `object`
        """
        for return_type, return_value in generator:
            if return_type == RETURN_TYPE_VALUE:
                return return_value
            
            self._capture(return_type, return_value)
        
        return default
    
    
    def _capture(self, return_type, return_value):
        """
        Captures an error / warning output.
        
        Parameters
        ----------
        return_type : `int`
            The type of the error.
        error_message : `str`
            Error message to capture.
        """
        captured = self._captured
        if captured is None:
            captured = []
            self._captured = captured
        
        captured.append((return_type, return_value))
    
    
    def _get_aggregated_state(self):
        """
        Gets the aggregated state of the captured errors.
        
        Returns
        -------
        error_type : `int`
        error_message : `None`, `str`
        """
        captured = self._captured
        if captured is None:
            return RETURN_TYPE_VALUE, None
        
        warning_messages = None
        exception_messages = None
        
        for return_type, return_value in captured:
            if return_type == RETURN_TYPE_WARNING:
                if warning_messages is None:
                    warning_messages = []
                
                warning_messages.append(return_value)
            
            else:
                if exception_messages is None:
                    exception_messages = []
                
                exception_messages.append(return_value)
        
        if warning_messages is None:
            if exception_messages is None:
                # should never happen
                error_type = RETURN_TYPE_VALUE
                error_message = None
            
            else:
                error_type = RETURN_TYPE_EXCEPTION
                if len(exception_messages) == 1:
                    error_message = exception_messages[0]
                
                else:
                    error_message = _build_error_message_single_source(exception_messages, 'exceptions')
        
        else:
            if exception_messages is None:
                error_type = RETURN_TYPE_WARNING
                
                if len(warning_messages) == 1:
                    error_message = warning_messages[0]
                
                else:
                    error_message =  _build_error_message_single_source(warning_messages, 'warnings')
            
            else:
                error_type = RETURN_TYPE_EXCEPTION
                error_message = _build_error_message_all_sources(exception_messages, warning_messages)
                

        return error_type, error_message


def _build_error_message_single_source(messages, source):
    """
    Builds error message for a single source.
    
    Parameters
    ----------
    messages : `list<str>`
        Collection of error messages.
    source : `str`
        The source of the error messages.
    
    Returns
    -------
    message : `str`
    """
    message_parts = ['Occurred ', source, ' while getting environmental variables (', str(len(messages)), '):']
    
    for message in messages:
        message_parts.append('\n')
        message_parts.append(message)
    
    return ''.join(message_parts)


def _build_error_message_all_sources(exception_messages, warning_messages):
    """
    Builds error message for a single source.
    
    Parameters
    ----------
    exception_messages : `list<str>`
        Collection of exception messages.
    warning_messages : `list<str>`
        Collection of warning messages.
    
    Returns
    -------
    message : `str`
    """
    message_parts = ['Occurred exceptions while getting environmental variables (', str(len(exception_messages)), '):']
    
    for message in exception_messages:
        message_parts.append('\n')
        message_parts.append(message)
    
    message_parts.append('\n\nAdditional warnings (')
    message_parts.append(str(len(warning_messages)))
    message_parts.append('):')
    
    for message in warning_messages:
        message_parts.append('\n')
        message_parts.append(message)
    
    return ''.join(message_parts)



__all__ = ('CONVERSION_FLAGS',)

from ...message import MessageFlag

from ....builder.constants import CONVERSION_KIND_FIELD
from ....builder.conversion import Conversion


class CONVERSION_FLAGS(Conversion):
    # Generic
    
    name = 'flags'
    name_aliases = None
    expected_types_messages = f'`None`, `int`, `{MessageFlag.__name__}`'
    kind = CONVERSION_KIND_FIELD
    output_conversion = None
    
    # Setting
    
    def set_merger(old_value, new_value):
        return old_value | new_value
    
    
    set_type = MessageFlag
    set_type_processor = None
    set_identifier = None
    set_listing_identifier = None
    
    
    def set_validator(value):
        # None
        if value is None:
            yield 0
            return
        
        # int | MessageFlag
        if isinstance(value, int):
            yield value
            return
        
        # No other cases
        return
    
    # Reading
    
    get_default = MessageFlag()
    
    
    def get_processor(value):
        return MessageFlag(value)
    
    
    # Serialization
    
    serializer_key = 'flags'
    
    
    def serializer_optional(value):
        if value:
            yield int(value)
    
    
    def serializer_required(value):
        return int(value)
    
    
    # Sorting
    
    sort_priority = 1800



__all__ = ()

from ...color import Color
from ...field_parsers import (
    bool_parser_factory, entity_id_parser_factory, flag_parser_factory, force_string_parser_factory, int_parser_factory,
    nullable_functional_parser_factory
)
from ...field_putters import (
    bool_optional_putter_factory, entity_id_putter_factory, flag_optional_putter_factory, force_string_putter_factory,
    int_putter_factory, nullable_functional_optional_putter_factory, string_flag_putter_factory
)
from ...field_validators import (
    bool_validator_factory, default_entity_validator_factory, entity_id_validator_factory, flag_validator_factory,
    force_string_validator_factory, int_conditional_validator_factory, nullable_entity_conditional_validator_factory,
    preinstanced_validator_factory
)
from ...permission import Permission
from ...permission.constants import PERMISSION_KEY

from ..role_manager_metadata import RoleManagerMetadataBase
from ..role_manager_metadata.constants import (
    APPLICATION_ROLE_CONNECTION_KEY, BOOSTER_KEY, BOT_ID_KEY, INTEGRATION_ID_KEY, SUBSCRIPTION_LISTING_ID_KEY
)

from .constants import NAME_LENGTH_MAX, NAME_LENGTH_MIN, ROLE_MANAGER_DEFAULT
from .flags import RoleFlag
from .preinstanced import RoleManagerType

# color

parse_color = flag_parser_factory('color', Color)
put_color = flag_optional_putter_factory('color', Color())
validate_color = flag_validator_factory('color', Color)

# flags

parse_flags = flag_parser_factory('flags', RoleFlag)
put_flags = flag_optional_putter_factory('flags', RoleFlag())
validate_flags = flag_validator_factory('flags', RoleFlag)

# guild_id

validate_guild_id = entity_id_validator_factory('guild_id', NotImplemented, include = 'Guild')

# id

parse_id = entity_id_parser_factory('id')
put_id = entity_id_putter_factory('id')
validate_id = entity_id_validator_factory('role_id')

# manager_metadata

validate_manager_metadata = default_entity_validator_factory(
    'manager_metadata', RoleManagerMetadataBase, default = ROLE_MANAGER_DEFAULT
)

# manager_type

validate_manager_type = preinstanced_validator_factory('manager_type', RoleManagerType)

# manager type & metadata

def parse_manager(data):
    """
    Parses out ``Role.manager_id`` and ``Role.manager_type` fields from the given data.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        Role data.
    
    Returns
    -------
    manager_type : ``RoleManagerType``
        If the role is manager, defined by what type of entity is the role is managed.
    
    manager_metadata : ``RoleManagerMetadataBase``
         Metadata about the role's manager.
    """
    managed = data.get('managed', False)
    manager_data = data.get('tags', None)
    
    while True:
        if (not managed):
            manager_type = RoleManagerType.none
            break
        
        if (manager_data is None) or (not manager_data):
            manager_type = RoleManagerType.unset
            break
        
        if BOT_ID_KEY in manager_data:
            manager_type = RoleManagerType.bot
            break
        
        if BOOSTER_KEY in manager_data:
            manager_type = RoleManagerType.booster
            break
        
        if SUBSCRIPTION_LISTING_ID_KEY in manager_data:
            manager_type = RoleManagerType.subscription
            break
        
        if APPLICATION_ROLE_CONNECTION_KEY in manager_data:
            manager_type = RoleManagerType.application_role_connection
            break
        
        if INTEGRATION_ID_KEY in manager_data:
            manager_type = RoleManagerType.integration
            break
        
        manager_type = RoleManagerType.unknown
        break
    
    metadata_type = manager_type.metadata_type
    if metadata_type is RoleManagerMetadataBase:
        manager_metadata = ROLE_MANAGER_DEFAULT
    else:
        manager_metadata = metadata_type.from_data(manager_data)
    
    return manager_type, manager_metadata


def put_manager(manager, data, defaults):
    """
    Puts the role's manager into the given data.
    
    Parameters
    ----------
    manager : `tuple` (``RoleManagerType``, ``RoleManagerMetadataBase``)
        The role's manager as a tuple containing it's identifier and type.
    
    data : `dict` of (`str`, `object`) items
        Role data.
    
    defaults : `bool`
        Whether default field values should be put into `data` as well.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    manager_type, manager_metadata = manager
    
    managed = manager_type is not RoleManagerType.none
    if defaults or managed:
        data['managed'] = managed
        data['tags'] = manager_metadata.to_data(defaults = defaults)
    
    return data


def validate_manager(manager):
    """
    Validates the given manager value.
    
    Parameters
    ----------
    manager : `None`, `tuple` (``RoleManagerType``, ``RoleManagerMetadataBase``)
        Role manager tuple containing it's identifier and type.
    
    Returns
    -------
    manager_type : ``RoleManagerType``
        If the role is manager, defined by what type of entity is the role is managed.
    
    manager_id : `int`
         If the role is managed, then it's manager's id if applicable.
    
    Raises
    ------
    TypeError
        - If `manager`'s type is invalid.
        - `manager_type` - `manager_metadata` mismatch.
    """
    if manager is None:
        return RoleManagerType.none, ROLE_MANAGER_DEFAULT
    
    if (not isinstance(manager, tuple)) or (len(manager) != 2):
        raise TypeError(
            f'`manager` can be `None` or `tuple` (of length 2 as manager_type & manager_metadata), got '
            f'{manager.__class__.__name__}; {manager!r}.'
        )
    
    manager_type, manager_metadata = manager
    manager_type = validate_manager_type(manager_type)
    manager_metadata = validate_manager_metadata(manager_metadata)
    
    expected_manager_metadata_type = manager_type.metadata_type
    received_manager_metadata_type = type(manager_metadata)
    
    if received_manager_metadata_type is RoleManagerMetadataBase:
        if expected_manager_metadata_type is not RoleManagerMetadataBase:
            manager_metadata = expected_manager_metadata_type()
    
    else:
        if received_manager_metadata_type is not expected_manager_metadata_type:
            raise TypeError(
                f'`manager_metadata` should be `{expected_manager_metadata_type.__name__}` instance, got '
                f'{received_manager_metadata_type.__name__}, {manager_metadata!r}.'
            )
    
    return manager_type, manager_metadata


# mentionable

parse_mentionable = bool_parser_factory('mentionable', False)
put_mentionable = bool_optional_putter_factory('mentionable', False)
validate_mentionable = bool_validator_factory('mentionable', False)

# name

parse_name = force_string_parser_factory('name')
put_name = force_string_putter_factory('name')
validate_name = force_string_validator_factory('name', NAME_LENGTH_MIN, NAME_LENGTH_MAX)

# permissions

parse_permissions = flag_parser_factory(PERMISSION_KEY, Permission)
put_permissions = string_flag_putter_factory(PERMISSION_KEY)
validate_permissions = flag_validator_factory('permission', Permission)

# position

parse_position = int_parser_factory('position', 0)
put_position = int_putter_factory('position')
validate_position = int_conditional_validator_factory(
    'position',
    0,
    (lambda position : position >= 0),
    '>= 0',
)

# separated

parse_separated = bool_parser_factory('hoist', False)
put_separated = bool_optional_putter_factory('hoist', False)
validate_separated = bool_validator_factory('separated', False)

# unicode_emoji

parse_unicode_emoji = nullable_functional_parser_factory(
    'unicode_emoji', NotImplemented, include = 'create_unicode_emoji'
)
put_unicode_emoji = nullable_functional_optional_putter_factory('unicode_emoji', lambda emoji: emoji.unicode)
validate_unicode_emoji = nullable_entity_conditional_validator_factory(
    'unicode_emoji', NotImplemented, lambda emoji: emoji.is_unicode_emoji(), 'unicode emoji', include = 'Emoji'
)
